# frozen_string_literal: true
require 'json'
module Hermes
  module Api
    module Dmo
      # checker Api::Dmo::HermesController avant d'ajouter des methodes
      class LessorsController < Hermes::Api::Dmo::ApplicationController
        set_instance_for :objectives

        VALID_OPERATORS = %w[
          $in
          $nin
          $ne
          $gt
          $gte
          $lt
          $lte
          $eq
          $regex
          $all
        ]

        ACCEPTABLE_PARAMS = %w[
          ids
          als_id
          created_at
          updated_at
          title
          siret
          siren
          is_root
          root_id
          parent_id
          id_rex
          id_cristal
          api_id
          rpls_id
          affiliate
          housing_offer_ids
          lessor_objective_ids
          group_id
          supervisor_group_id
          manager_group_id
        ]

        REGEX_PARAMS = %w[
          title
          siret
          siren
        ]

        OBJECTIVES_ACCEPTABLE_PARAMS = %w[
          created_at
          updated_at
          activation_date
          department
          district
          threshold_type
          priority_public
          new_offer
          target_by_typologies
          realized_by_typologies
          present_typologies
          agency_id
        ]

        OBJECTIVES_REGEX_PARAMS = %w[
          district
        ]

        MUTATE_PARAMS = %i[
          title
          siret
          siren
          id_rex
          id_cristal
          api_id
          rpls_id
          affiliate
          housing_offer_ids
          lessor_objective_ids
        ]
        def mutate_params
          MUTATE_PARAMS
        end 

        def index
          acceptable_params *ACCEPTABLE_PARAMS
          regex_params *REGEX_PARAMS
          valid_operators *VALID_OPERATORS
          super
        end

        def objective_handle_params
          acceptable_params *OBJECTIVES_ACCEPTABLE_PARAMS
          regex_params *OBJECTIVES_REGEX_PARAMS
          valid_operators *VALID_OPERATORS
        end
        
        def admin_search
          return render json: {status: "unauthorized", code: 401, message: "user is no superadmin"} unless is_super_admin

          lessor_query = admin_search_param(params['lessor_params'], 'lessor', [:id, :id_cristal, :siren, :siret], [], [:title])
          user_query = admin_search_param(params['user_params'], 'user', [:id, :email, :auth_roles], [:first_name, :last_name])

          query = [
            {
               "$lookup": {
                   from: 'hermes_entities',
                   localField: "lessor_id",
                   foreignField: "_id",
                   as: "lessor"
               }
            }, 
            { "$unwind": { path: "$lessor", preserveNullAndEmptyArrays: true } },
            { "$match": lessor_query },
            {
              "$lookup":{
                from: 'users',
                localField: "user_id",
                foreignField: "_id",
                as: "user"
              }
            },
            { "$unwind": { path: "$user", preserveNullAndEmptyArrays: true } },
            { "$match": user_query },
          ]

          aggregation = Hermes::UserRight.collection.aggregate(query.push(facet_count_query(params)))

          mapped_attributes = {
            right: "['right']",
            user_id: "['user_id']",
            user_email: "['user']['email']",
            user_first_name: "['user']['first_name']",
            user_last_name: "['user']['last_name']",
            user_auth_roles: "['user']['auth_roles']",
            user_phone_number: "['user']['phone_number']",
            lessor_id: "['lessor_id']",
            lessor_als_id: "['lessor']['als_id']",
            lessor_title: "['lessor']['title']",
            lessor_siret: "['lessor']['siret']",
            lessor_siren: "['lessor']['siren']",
            lessor_id_cristal: "['lessor']['id_cristal']",
          }

          paginate_like_array(aggregation, mapped_attributes)
        end

        def paginate_objectives(objectives)
          objective_handle_params
          objectives = objectives.where(build_query(Hermes::LessorObjective))
                      .order(order_params)
          Rails.logger.debug "TEST #{objectives.selector}"
          paginate(
            objectives,
            each_serializer: Hermes::LessorObjectiveSerializer
          )
        end

        def all_objectives
          objectives = @instance.lessor_objectives
          paginate_objectives(objectives)
        end

        def active_objectives
          objectives = @instance.lessor_objectives.actives
          paginate_objectives(objectives)
        end

        def switch_manual_mode
          if is_super_admin
            lessor = Hermes::Lessor.find(params[:id])
            if lessor.switch_manual_mode
              render json: {status: "success", code: 200, message: "manual mode switch to #{lessor.manual_mode}"}
            else
              render json: {status: "not found", code: 404, message: 'lessor not found'}
            end 
          else 
            render json: {status: "unauthorized", code: 401, message: "user is not a superadmin"}
          end 
        end 

        def current_operation_names
          operation_names = @instance.current_operation_names
          render :json => operation_names
        end

        def add_user
          if @instance.writable_by? current_user
            user = ::User.find params[:data][:attributes][:user_id]
            if user && user.update(lessor_id: @instance.id) 
              return render json: { data: 'user updated' }
            end
          end
          
          render_json_error(:not_found, :object_not_found) unless @instance
        end

        def remove_user
          if @instance.writable_by? current_user
            user = ::User.find params[:data][:attributes][:user_id]
            if user && user.update(lessor_id: nil) 
              return render json: { data: 'user updated' }
            end
          end
          
          render_json_error(:not_found, :object_not_found) unless @instance
        end

        def rdsl
          return render_json_error(:not_found, :object_not_found) unless @instance

          user_ids = Hermes::UserRight.where(lessor_id: @instance.id, right: 'administrator').pluck(:user_id)
          users = User.where(:id.in => user_ids, auth_roles: 'rdsl')

          paginate(
            users,
            each_serializer: Hermes::UserSerializer,
            **@options
          )
        end
      end
    end
  end
end
