# frozen_string_literal: true

module Hermes
  # Dossier de candidature , gere le listing des pieces a fournir et le statut du dossier
  class HousingRequestFolder
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::SopraXmlSender
    include Hermes::Concerns::ChangeMinusTwoYearConcern
    
    field :nud_status, type: String
    field :consultation_status, type: String
    field :application_status, type: String
    
    field :completion, type: Integer # pourcentage de completion du dossier
    field :file_upload_from_sne_complete, type: Boolean
    belongs_to :housing_request, class_name: 'Hermes::HousingRequest', index: true
    has_and_belongs_to_many :required_documents, class_name: 'Hermes::RequiredDocument', inverse_of: nil
    
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    writer_auth_roles 'collaborateur entreprise', "collaborateur offre logement", 'superadministrateur'
    
    common_indexes *%i[
      nud_status
      consultation_status
      application_status
      completion
      file_upload_from_sne_complete
    ]
    
    before_create :before_create_action
    before_save :before_save_action
    after_update :after_update_action
    # METODES DE CREATION DES DOCUENTS 5e4a5ad322ec7927f7dc62eb 5e4a5ad322ec7927f7dc62ea
    def compute_required_documents
      self.required_document_ids = []
      find_or_create_documents
    end

    def get_types
      required_documents.pluck(:type)
    end 

    def types_include(types)
      types = [types] unless types.kind_of?(Array)
      (types - get_types) == []
    end
    
    def get_labels
      labels = []
      required_documents.each do |doc|
        
        labels.push(id: doc.id,
                    type: doc.type,
                    applicant_id: doc.applicant_id,
                    status: doc.status,
                    declared: doc.declared,
                    rf_nud: (doc.folder_where_is_required_for_nud.include? id),
                    rf_consultation: (doc.folder_where_is_required_for_consultation.include? id),
                    rf_appli: (doc.folder_where_is_required_for_application.include? id),
                    ihm_order: doc.ihm_order
        )
      
      end
      labels
    end
    
    def get_not_declared
      required_documents.where(:declared.ne => true).pluck(:type)
    end 

    
    def all_applicant_ids
      all_applicant_ids = []
      [housing_request.applicant_ids, housing_request.main_applicant_id].flatten.uniq.compact.each do |id|
        all_applicant_ids.push id.to_s
      end
      all_applicant_ids
    end
    
    def find_or_create_documents
      compute_identity_documents(housing_request.main_applicant_id.to_s, true)
      #      && Rails.env == 'test')
      housing_request.applicant_ids&.each do |applicant_id|
        compute_identity_documents(applicant_id.to_s, false)
      end
      
      
      # required_documents_common_type_list = ['hermes_identite', 'hermes_revenu_fiscal', 'hermes_ressources_mensuelles']
      all_applicant_ids.each do |applicant_id|
        compute_ressources_documents(applicant_id, applicant_id == housing_request.main_applicant_id.to_s, housing_request.is_untensed_area)
        compute_revenus_fiscaux(applicant_id, applicant_id == housing_request.main_applicant_id.to_s, housing_request.is_untensed_area)
        compute_dependents_documents(applicant_id, applicant_id == housing_request.main_applicant_id.to_s, housing_request.is_untensed_area)
        compute_family_situation_documents(applicant_id, applicant_id == housing_request.main_applicant_id.to_s, housing_request.is_untensed_area)
      end

      
      #   compute_others_documents_for_main_applicant(housing_request.main_applicant, true)
      compute_actual_housing_documents(housing_request.main_applicant_id.to_s, true, housing_request.is_untensed_area)
      compute_motivation_documents(housing_request.main_applicant_id.to_s, true, housing_request.is_untensed_area)
      compute_disabilities_documents(housing_request.main_applicant_id.to_s, true, housing_request.is_untensed_area)
    end
    
    def compute_identity_documents(applicant_id, is_main_applicant)
      applicant = Applicant.find(applicant_id)
      if applicant
        doc = case applicant.nationality
              when '1'
                {ihm_order: is_main_applicant ? 'A01' : 'A02',
                 type_ihm: "CNI",
                 type: "Carte nationale d’identité ou passeport",
                 message: "Veuillez déposer une copie recto-verso de votre carte d'identité ou en cours de validité. 
Attention, ce document doit impérativement faire figurer 
Nom d’usage, Nom de naissance, Prénom (uniquement le premier),Date de naissance,Lieu de naissance,Pays de naissance, Date d’émission, Date d’expiration, Signature et photo. 
Ce document doit être lisible et complet.",
                 code: is_main_applicant ? 'CID' : 'CICO',
                 category: is_main_applicant ? 'IDD' : 'IDCO',
                 label: is_main_applicant ? "Carte d'identité ou passeport du demandeur" : "Carte d'identité ou passeport du conjoint ou du co-titulaire",
                }
              when '3'
                {ihm_order: is_main_applicant ? 'A01' : 'A02', #A03 lorsque on dédoublera pour le récépissé todo
                 type_ihm: "ID_ETR",
                 type: "titre de séjour et Récépissé de renouvellement de titre de séjour si celui ci est expiré",
                 message: "Veuillez déposer une copie de votre titre de séjour en cours de validité. 
Attention, ce document doit impérativement faire figurer 
Civilité, Nom d’usage, Nom de naissance, Prénom (uniquement le premier),Date de naissance,Lieu de naissance,Pays de naissance, Date d’émission, Date d’expiration, Signature et photo.
Ce document doit être lisible et complet. 
Si le titre de séjour est expiré, veuillez déposer le récépissé du renouvellement de votre titre de séjour datant de moins de 6 mois",
                 code: is_main_applicant ? 'TSD' : 'TSCO',
                 category: is_main_applicant ? 'IDD' : 'IDCO',
                 label: is_main_applicant ? 'Titre de séjour du demandeur' : 'Titre de séjour du conjoint ou du co-titulaire',
                }
              when '2'
                {ihm_order: is_main_applicant ? 'A01' : 'A02',
                 type_ihm: "CNI_UE",
                 type: "Carte d'identité d'un pays européen",
                 message: "Veuillez déposer une copie recto-verso de votre carte d'identité ou passeport en cours de validité. 
Ce document doit être lisible et complet. ",
                 code: is_main_applicant ? 'CID' : 'CICO',
                 category: is_main_applicant ? 'IDD' : 'IDCO',
                 label: is_main_applicant ? "Carte d'identité ou passeport du demandeur" : "Carte d'identité ou passeport du conjoint ou du co-titulaire",
                 }
              end
        if doc
          
          
          find_or_create_required_document(
            doc[:ihm_order],
            "Pièce d'identité",
            doc[:type],
            doc[:message],
            applicant.id.to_s,
            is_main_applicant,
            is_main_applicant,
            true,
            'SPI',
            doc[:code],
            doc[:category],
            nil, doc[:label],
            doc[:type_ihm])
        
        end
      end
    end
    
    def get_month_label(delay)
      date = (Date.current) + (delay * 31).send("days")
      month_labels = ['de Janvier', 'de Février', 'de Mars', "d'Avril", 'de Mai', 'de Juin', 'de Juillet', "d'Août", 'de Septembre', "d'Octobre", 'de Novembre', 'de Décembre']
      month_labels[date.month - 1] + " " + date.year.to_s
    end
    
    def compute_ressources_documents(applicant_id, _is_main_applicant, is_untensed_area)
      applicant = Applicant.find(applicant_id)
      return unless applicant
      
      rules = [
        {
          monthly_incomes: %w[AUT],
          required_documents: [{
                                 ihm_order: "BXX",
                                 type: 'Justificatif(s) autres ressources perçues',
                                 message: "Veuillez déposer ici un Justificatif des  autres ressources perçues. Ce document doit être lisible et complet. ",
                                 code: 'JUSR',
                                 category: 'RESS',
                                 label: 'Justificatif de ressources'}],
          
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[BE],
          required_documents: [{
                                 ihm_order: "B09",
                                 type: "Carte d'étudiant délivrée par un établissement certifié et valable pour l'année en cours",
                                 message: "Veuillez déposer ici une copie recto verso de votre carte d'étudiant ou une attestation scolaire pour l'année en cours établie par un établissement certifié.Ce document doit être lisible et complet. ",
                                 code: 'CAE',
                                 category: 'SITPRO',
                                 label: "Carte d'étudiant"
                               },
                               {
                                 ihm_order: "B09",
                                 type: "Avis d'attribution de bourse",
                                 message: "Veuillez déposer ici une copie de votre avis d'attribution de bourse CROUS. Attention, ce document doit impérativement faire figurer votre nom, prénom, ainsi que les montants perçus. Ce document doit être lisible et complet. "
                               }],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[RET],
          required_documents: [{ihm_order: "B08",
                                type: "Notification de pension",
                                message: "Veuillez déposer ici une copie de votre attestation de retraite. 
Vous pouvez la retrouver sur https://www.lassuranceretraite.fr/portail-info/home.html. 
Attention, ce document doit impérativement faire figurer votre régime de retraite et les montants perçus. 
Ce document doit être lisible et complet.",
                                code: 'JUSR',
                                category: 'RESS',
                                label: 'Justificatif de ressources'}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[PINV],
          #todo===> a verifier
          required_documents: [{ihm_order: "B05",
                                type: 'Attestation de la Caisse d’allocations familiales (CAF) /Mutualité sociale agricole (MSA)',
                                message: "Veuillez déposer ici une copie de l'attestation de la CAF OU de la MSA si vous faites partie du secteur agricole. 
Attention, ce document doit impérativement faire figurer nom, prénom et montants perçus. 
Ce document doit être lisible et complet. ",
                                code: 'JUSR',
                                category: 'RESS',
                                label: 'Justificatif de ressources'}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[SAL],
          work_contract_type: %w[APP],
          required_documents: [{ihm_order: "B04",
                                type: "Contrat de travail de l'année en cours/ Bulletins de salaire du mois" + get_month_label(-1),
                                message: "Veuillez déposer ici votre bulletins de salaire du mois" + get_month_label(-1) + " ou votre contrat d'apprentissage/ professionnalisation pour l'année en cours. Ce document doit être lisible et complet. ",
                                code: 'CONT',
                                category: 'SITPRO',
                                label: 'Contrat de travail'},
                               {ihm_order: "B04",
                                type: "Contrat de travail de l'année en cours/ Bulletins de salaire du mois" + get_month_label(-2),
                                message: "Veuillez déposer ici votre bulletins de salaire du mois" + get_month_label(-2) + " ou votre contrat d'apprentissage/ professionnalisation pour l'année en cours. Ce document doit être lisible et complet. ",
                                code: 'CONT',
                                category: 'SITPRO',
                                label: 'Contrat de travail'},
                               {ihm_order: "B04",
                                type: "Contrat de travail de l'année en cours/ Bulletins de salaire du mois" + get_month_label(-3),
                                message: "Veuillez déposer ici votre bulletins de salaire du mois" + get_month_label(-3) + " ou votre contrat d'apprentissage/ professionnalisation pour l'année en cours. Ce document doit être lisible et complet. ",
                                code: 'CONT',
                                category: 'SITPRO',
                                label: 'Contrat de travail'},
          ],
          r_for_consultation: _is_main_applicant && !is_untensed_area,
          r_for_application: true
        }, {
          monthly_incomes: %w[SAL],
          work_contract_type: %w[AUT],
          #todo===>souci=
          required_documents: [{ihm_order: "BXX",
                                type: 'justificatif(s) du salaire perçu',
                                message: "Veuillez déposer ici un Justificatif des salaires perçus. Ce document doit être lisible et complet. ", code: 'ATTF', category: 'SITPRO', label: 'Attestation de formation ou autre justificatif de situation'}],
          r_for_consultation: _is_main_applicant && !is_untensed_area,
          r_for_application: true
        }, {
          monthly_incomes: %w[SAL],
          work_contract_type: %w[CDI CDD CDIP AGP],
          
          required_documents: [{ihm_order: "B04",
                                type: "Bulletin de salaire privé / public du mois " + get_month_label(-1) + " ou Attestation de l'employeur",
                                message: "Veuillez déposer ici votre bulletin de salaire du mois de " + get_month_label(-1) + ". 
Attention, ces documents doivent impérativement faire figurer  Nom d’usage, Nom de naissance, Prénom, SIRET, Salaire net mensuel perçu, Date du document, Prélèvement à la source si simple ?
Ce document doit être lisible et complet. 
ou
Veuillez déposer ici une attestation de votre employeur datant de moins de 3 mois. Ce document doit être lisible et complet.",
                                code: 'FPD',
                                category: 'RESS',
                                label: 'Fiche de paie du demandeur et le cas échéant de son conjoint et de chaque co-titulaire',
                                type_ihm: "3_SALAIR"},
                               {ihm_order: "B04",
                                type: "Bulletin de salaire privé / public du mois " + get_month_label(-2) + " ou Attestation de l'employeur",
                                message: "Veuillez déposer ici votre bulletin de salaire du mois de " + get_month_label(-3) + ". 
Attention, ces documents doivent impérativement faire figurer  Nom d’usage, Nom de naissance, Prénom, SIRET, Salaire net mensuel perçu, Date du document, Prélèvement à la source si simple ?
Ce document doit être lisible et complet. 
ou
Veuillez déposer ici une attestation de votre employeur datant de moins de 3 mois. Ce document doit être lisible et complet.",
                                code: 'FPD',
                                category: 'RESS',
                                label: 'Fiche de paie du demandeur et le cas échéant de son conjoint et de chaque co-titulaire',
                                type_ihm: "3_SALAIR"},
                               {ihm_order: "B04",
                                type: "Bulletin de salaire privé / public du mois " + get_month_label(-3) + " ou Attestation de l'employeur",
                                message: "Veuillez déposer ici votre bulletin de salaire du mois de " + get_month_label(-3) + ". 
Attention, ces documents doivent impérativement faire figurer  Nom d’usage, Nom de naissance, Prénom, SIRET, Salaire net mensuel perçu, Date du document, Prélèvement à la source si simple ?
Ce document doit être lisible et complet. 
ou
Veuillez déposer ici une attestation de votre employeur datant de moins de 3 mois. Ce document doit être lisible et complet.",
                                code: 'FPD',
                                category: 'RESS',
                                label: 'Fiche de paie du demandeur et le cas échéant de son conjoint et de chaque co-titulaire',
                                type_ihm: "3_SALAIR"}],
          
          
          r_for_consultation: _is_main_applicant && !is_untensed_area,
          r_for_application: true
        }, {
          monthly_incomes: %w[SAL],
          work_contract_type: %w[ART],
          required_documents: [{ihm_order: "BXX",
                                type: "Dernier bilan ou attestation du comptable de l'entreprise évaluant le salaire mensuel perçu ou tout document comptable habituellement fourni à l'administration permettant de justifier les revenus",
                                message: "Veuillez déposer ici un bilan ou une attestation comptable de l'entreprise ou tout document vous permettant de justifier vos revenus datant de moins de 3 mois. Attention, ce document doit impérativement faire figurer votre SIRET et le salaire net mensuel perçu.
Ce document doit être lisible et complet.",
                                code: 'JUSR', category: 'RESS', label: 'Justificatif de ressources'}],
          r_for_consultation: _is_main_applicant && !is_untensed_area,
          r_for_application: true
        }, {
          monthly_incomes: %w[AMV],
          #todo===>pas trouvé
          required_documents: [{ihm_order: "BXX", type: "Notification de pension de minimum vieillesse",
                                message: 'Veuillez déposer ici la notification de pension de minimum vieillesse.
Ce document doit être lisible et complet. ',
                                code: 'JUSR',
                                category: 'RESS',
                                label: 'Justificatif de ressources'}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[PAR],
          required_documents: [{ihm_order: "BXX",
                                type: "Extrait du jugement de divorce ou autre document démontrant la perception de la pension",
                                message: "Veuillez déposer ici une copie de l'extrait du jugement attestant du montant perçu ou, dans le cas d'un divorce à l'amiable, une attestation sur l'honneur. 
Ce document doit être lisible et complet.",
                                code: 'JUSR',
                                category: 'RESS',
                                label: 'Justificatif de ressources'}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[CHO],
          required_documents: [{ihm_order: "B07",
                                type: "Avis de paiement de l'ARE",
                                message: 'Veuillez déposer ici une copie de votre avis de paiement Pôle Emploi en cours de validité. 
Vous pouvez retrouver ce document sur votre espace personnel Pôle Emploi. 
Attention, ce document doit impérativement faire figurer les montants perçus. 
Ce document doit être lisible et complet. ',
                                code: 'JUSR',
                                category: 'RESS',
                                label: 'Justificatif de ressources'}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[
            AF
            AAH
            AEEH
            AJPP
            PAJE
            PDA
          ],
          required_documents: [{ihm_order: "B05",
                                type: "Attestation de la Caisse d’allocations familiales (CAF) /Mutualité sociale agricole (MSA)",
                                message: "Veuillez déposer ici une copie de l'attestation de la CAF OU de la MSA si vous faites partie du secteur agricole. 
Attention, ce document doit impérativement faire figurer nom, prénom et montants perçus. 
Ce document doit être lisible et complet. ",
                                code: 'CAF',
                                category: 'REV',
                                label: "Certificat de l'administration fiscale"}],
          r_for_consultation: false,
          r_for_application: true
        }, {
          monthly_incomes: %w[
            RSA
          ],
          required_documents: [{ihm_order: "B05",
                                type: "Attestation de la Caisse d’allocations familiales (CAF) /Mutualité sociale agricole (MSA)",
                                message: "Veuillez déposer ici une copie de votre attestation RSA en cours de validité. Pour accéder à votre attestation de RSA, nous vous invitons à vous connecter à votre espace personnel sur le site Caf.fr , rubrique \"Mes attestations\". Si vous faites partie du secteur agricole, vous pouvez retrouver ce document sur votre espace personnel MSA. Ce document doit être lisible et complet.",
                                code: 'CAF',
                                category: 'REV',
                                label: "Certificat de l'administration fiscale"}],
          r_for_consultation: false,
          r_for_application: true
        }
      ]
      if applicant&.monthly_charges && ((applicant&.monthly_charges["PAV"]&.to_i ||
        applicant.monthly_charges["PAV".to_sym]&.to_i ||
        0
      ) > 0)
        find_or_create_required_document(
          "BXX",
          "Ressources",
          "Extrait de jugement ou autre document démontrant le versement de la pension",
          "Veuillez déposer ici une copie de l'extrait du jugement attestant du montant versé ou, dans le cas d'un divorce à l'amiable, une attestation sur l'honneur. Ce document doit être lisible et complet. ",
          applicant.id.to_s,
          false,
          false,
          true,
          'SPR',
          "JUSR",
          "RESS",
          nil,
          'Justificatif de ressources',
          false)
      end
      rules.each do |rule|
        work_contract_type_condition = true
        if rule[:work_contract_type]
          work_contract_type_condition = rule[:work_contract_type].include? applicant.work_contract_type
        end
        is_doc_needed = false
        rule[:monthly_incomes].each do |ressource_label|
          next unless applicant.monthly_incomes &&
            (
            applicant.monthly_incomes[ressource_label]&.to_i ||
              applicant.monthly_incomes[ressource_label.to_sym]&.to_i ||
              0
            ) > 0 && work_contract_type_condition
          
          is_doc_needed = true
        end
        next unless is_doc_needed
        
        rule[:required_documents].each do |req_doc|
          find_or_create_required_document(req_doc[:ihm_order],
                                           "Ressources",
                                           req_doc[:type],
                                           req_doc[:message],
                                           applicant.id.to_s,
                                           false,
                                           rule[:r_for_consultation],
                                           rule[:r_for_application],
                                           'SPR', req_doc[:code],
                                           req_doc[:category],
                                           nil,
                                           req_doc[:label],
                                           req_doc[:type_ihm])
        
        end
      end
    end
    
    def compute_others_documents_for_main_applicant(applicant_id, _is_main_applicant, is_untensed_area)
      applicant = Applicant.find(applicant_id)
      return unless applicant
      
      if applicant.family_situation == 'V'
        find_or_create_required_document(
          #todo===> veuf pas trouvé
          "DXX",
          'Certificat de décès ou livret de famille',
          'Certificat de décès ou livret de famille',
          "Veuillez déposer ici le certificat de décès ou livret de famille. Ce document doit être lisible et complet. ",
          applicant.id.to_s,
          false,
          false,
          true,
          'SPR'
        )
      end
    end
    
    def get_year_label(delay)
      (get_minus_two_year.to_i + 2 - delay.to_i).to_s
    end

    def compute_revenus_fiscaux(applicant_id, is_main_applicant, is_untensed_area)
      applicant = Applicant.find(applicant_id)
      return unless applicant

      coapp = nil

      if !is_main_applicant && !housing_request.coapplicants.empty?
        coapp = housing_request.coapplicants.detect { |a| a && a[:id] == applicant_id.to_s }
      end

      not_need_rd = [
        'Conjoint',
        'Conjoint (e)',
        'M',
        'Pacsée',
        'P',
      ]

      if is_main_applicant || !applicant.dynamic_amount_minus_two.nil? || (coapp && !not_need_rd.include?(coapp[:rel_type]))
        find_or_create_required_document(
          is_main_applicant ? 'B01' : 'B02',
          "Avis d'imposition",
          ["Avis d'imposition "+get_year_label(1)+" sur les revenus de l'année "+get_year_label(2), "Avis d'imposition N-2"],
          "Copie intégrale (4 pages) de l'Avis d'imposition ou def non imposition de l'année "+get_year_label(2)+" ou avis de situation déclarative à l'impôt sur le revenu (Attention, ce document doit impérativement faire figurer
  Civilité,Nom d’usage, Nom de naissance, Prénom, Numéro fiscal du déclarant 1, Référence de l'avis, Revenu Fiscal de Référence , Année de l'avis d'imposition, Adresse, Code postal, Ville, Pays, Date du document et la composition familiale.)
  ou
  Copie intégrale de l'avis d'imposition ou de non imposition des parents si fiscalement rattaché. (ce document doit impérativement faire figurer
  Civilité ,Nom d’usage, Nom de naissance, Prénom, Numéro fiscal du déclarant 1
  ,Référence de l'avis, Année de l'avis d'imposition, Adresse, Code postal, Ville,Pays,Date du document ainsi que la rubrique C   \"personne à charge en xxxx ou D \"rattachement en xxxx d'enfants majeurs.)
  ou
  Imposition à l'étranger :
  Avis d'imposition sur le revenu du pays ou attestation d'une administration compétente (Ces documents doivent être traduits en français et convertis en euros) (Ce document doit impérativement faire figurer : l'avis d'imposition sur le revenu dans l'Etat ou le territoire collecteur si tout ou une partie de vos revenus perçus l’avant-dernière année (N–2) n’a pas été imposé en France mais dans un autre Etat ou territoire.
  Ces documents doivent être lisibles et complets.",
          applicant.id.to_s,
          false,
          !is_untensed_area,
          true,
          'RFR',
          is_main_applicant ? 'AVID' : 'AVIDCH',
          'REV',
          false,
          is_main_applicant ? "Avis d'imposition ou de non imposition (année N-2 et le cas échéant N-1) du demandeur et le cas
            échéant de son conjoint" : "Avis d'imposition ou de non imposition (année N-2 et le cas échéant N-1) du co-titulaire ou des
            personnes à charge",
          "IMPO_N2",
         
        )
      end 
      
      if applicant&.dynamic_amount_minus_one && applicant&.dynamic_amount_minus_one >= 0
        find_or_create_required_document(
          is_main_applicant ? 'B01' : 'B02',
          "Avis d'imposition",
          ["Avis d'imposition "+get_year_label(0)+" sur les revenus de l'année "+get_year_label(1), "Avis d'imposition N-1"] ,
          "Copie intégrale (4 pages) de l'Avis d'imposition ou de non imposition de l'année "+get_year_label(1)+" ou avis de situation déclarative à l'impôt sur le revenu (Attention, ce document doit impérativement faire figurer
Civilité,Nom d’usage, Nom de naissance, Prénom, Numéro fiscal du déclarant 1, Référence de l'avis, Revenu Fiscal de Référence , Année de l'avis d'imposition, Adresse, Code postal, Ville, Pays, Date du document et la composition familiale.)
ou
Copie intégrale de l'avis d'imposition ou de non imposition des parents si fiscalement rattaché. (ce document doit impérativement faire figurer
Civilité ,Nom d’usage, Nom de naissance, Prénom, Numéro fiscal du déclarant 1
,Référence de l'avis, Année de l'avis d'imposition, Adresse, Code postal, Ville,Pays,Date du document ainsi que la rubrique C   \"personne à charge en xxxx ou D \"rattachement en xxxx d'enfants majeurs.)
ou
Imposition à l'étranger :
Avis d'imposition sur le revenu du pays ou attestation d'une administration compétente (Ces documents doivent être traduits en français et convertis en euros) (Ce document doit impérativement faire figurer : l'avis d'imposition sur le revenu dans l'Etat ou le territoire collecteur si tout ou une partie de vos revenus perçus l’avant-dernière année (N–1) n’a pas été imposé en France mais dans un autre Etat ou territoire.
Ces documents doivent être lisibles et complets.",
          applicant.id.to_s,
          false,
          !is_untensed_area,
          true,
          'RFR',
          is_main_applicant ? 'AVID' : 'AVIDCH',
          'REV',
          false,
          is_main_applicant ? "Avis d'imposition ou de non imposition (année N-2 et le cas échéant N-1) du demandeur et le cas
            échéant de son conjoint" : "Avis d'imposition ou de non imposition (année N-2 et le cas échéant N-1) du co-titulaire ou des
            personnes à charge",
          "IMPO_N1",
        )
      end
    end
    
    def compute_family_situation_documents(applicant_id, is_main_applicant, is_untensed_area)
      applicant = Applicant.find(applicant_id)
      return unless applicant
      
      rules = [
        {
          situation: 'P',
          required_document: {ihm_order: "DXX",
                              type: 'Attestation d\'enregistrement de PACS',
                              message: "Veuillez déposer ici une copie du récépissé de l'enregistrement du PACS délivré par votre mairie/
 un duplicata du PACS si celui a été enregistré chez un notaire/ 
une copie de vos actes de naissance respectifs où il est fait mention de votre Pacs. 
Ce document doit être lisible et complet."},
          not_for_coap: true
        }, {
          situation: 'M',
          required_document: {ihm_order: "A05",
                              type: 'Livret de famille ou tout document équivalant démontrant le mariage',
                              message: "Veuillez déposer ici l'extrait du livret de famille ou tout document équivalent démontrant le mariage. Attention, les noms des deux conjoints doivent être visibles. Ce document doit être lisible et complet. ",
                              code: is_main_applicant ? nil : 'LFCO',
                              category: is_main_applicant ? nil : 'IDCO',
                              label: is_main_applicant ? nil : "Livret de famille ou acte d'État civil"},
          not_for_coap: true
        }, {
          situation: 'D',
          required_document: {ihm_order: "A07",
                              type: "Jugement de divorce ou convention homologuée en cas de divorce par consentement mutuel",
                              message: 'Veuillez déposer ici une copie du jugement de divorce OU la convention homologuée si vous avez divorcé par consentement mutuel. 
Ce document doit être lisible et complet.',
                              code: 'JUGFAM',
                              category: 'MOTIF',
                              label: 'Jugement de divorce, 
ordonnance de non conciliation ou autres jugements familiaux'}
        }, {
          situation: 'S',
          required_document: {ihm_order: "A07",
                              type: "Jugement de divorce ou convention homologuée en cas de divorce par consentement mutuel 
ou
Mention de dissolution du pacs dans l'acte de naissance
ou
Séparation d'un couple pacsé récépissé d'enregistrement de la déclaration
ou
Ordonnance de non-conciliation ou, à défaut, copie de l’acte de saisine du juge aux affaires familiales dans les conditions prévues au code de procédure civile ou, lorsque c’est un divorce par consentement mutuel, justificatif de l’avocat du demandeur",
                              message: "(Vous êtiez marié(e))
« Veuillez déposer ici une copie du jugement de divorce OU la convention homologuée si vous avez divorcé par consentement mutuel. Ce document doit être lisible et complet.  »
ou
(Vous êtiez pacsé(e))
« Veuillez déposer ici une copie de la mention de dissolution du PACS figurant sur votre acte de naissance. Ce document doit être lisible et complet. »
ou
(Autre)
« Récépissé du dépôt de plainte si violence au sein du couple »
"}
        }, {
          situation: 'V',
          #todo====>souci
          required_document: {ihm_order: "AXX",
                              type: "Certificat de décès ou livret de famille",
                              message: 'Certificat de décès ou livret de famille'}
        }
      
      ]
      
      rules.each do |rule|
        if rule[:situation] == applicant.family_situation
          (is_main_applicant && rule[:not_for_coap]) || !rule[:not_for_coap]
          find_or_create_required_document(
            rule[:required_document][:ihm_order],
            "Situation familiale",
            rule[:required_document][:type],
            rule[:required_document][:message],
            applicant.id.to_s,
            false,
            false,
            true, 'SF',
            rule[:required_document][:code],
            rule[:required_document][:category],
            false,
            rule[:required_document][:label])
        end
      end
    end
    
    def compute_actual_housing_documents(applicant_id, is_main_applicant, is_untensed_area)
      applicant = Applicant.find(applicant_id)
      return unless applicant
      
      rules = [
        {
          actual_housing_detail: %w[POCC],
          #todo =>>>attention, piece réunifiées redédoublées
          required_documents: [{ihm_order: "CXX",
                                type: "Acte de propriété",
                                message: 'Veuillez déposer ici une copie de votre acte de propriété. Ce document doit être lisible et complet. ',
                                code: 'COLOC',
                                category: 'LOGACT',
                                label: 'Contrat de location / justificatif de propriété'},
                               {ihm_order: "CXX",
                                type: "Plan de financement",
                                message: 'Veuillez déposer ici une copie de votre plan de financement en cours de validité. Ce document doit être lisible et complet. '
                               }]
        }, {
          actual_housing_detail: %w[RSLL HLF RHVS HSH HHSH HCDFCM CDEFCM RE REU APCT],
          required_documents: [{ihm_order: "C03",
                                type: "Attestation de la structure d’hébergement ou du gestionnaire du logement-foyer",
                                message: "Veuillez déposer ici une attestation de votre structure d'hébergement ou produite par le gestionnaire de votre logement-foyer datant de moins de 3 mois. 
Attention votre nom, celui de l'hébergeant ainsi que l'adresse doivent impérativement figurer sur ce document. 
Ce document doit être lisible et complet. "}]
        }, {
          #todo ===>absent de la nouvelle liste
          actual_housing_detail: %w[FONC],
          required_documents: [{ihm_order: "C01",
                                type: "Contrat de bail",
                                message: "Veuillez déposer ici une copie de votre contrat de bail. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour'},
                               {ihm_order: "C01",
                                type: "Quittance de loyer ou attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués",
                                message: "Veuillez déposer ici une copie de votre dernière quittance de loyer. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',}
=begin
                               {ihm_order: "C01",
                                type: "quittance de loyer",
                                message: "Veuillez déposer ici une copie de votre dernière quittance de loyer. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',
                                or: ["Attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués"]},

                               {ihm_order: "C01",
                                type: "Attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués",
                                message: "Veuillez déposer ici une attestation de votre bailleur datant de moins de 3 mois indiquant que votre situation locative est à jour. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',
                                or: ["quittance de loyer"]}
=end
          ]
        }, {
          #todo ===>absent de la nouvelle liste
          actual_housing_detail: %w[HCPE CPE HCP CP],
          required_documents: [{ihm_order: "C03",
                                type: 'Attestation de la personne qui héberge',
                                message: "Veuillez déposer ici une attestation datant de moins de 3 mois rédigée par la personne qui vous héberge. 
Attention votre nom, celui de l'hébergeant ainsi que l'adresse doivent impérativement figurer sur ce document.
Ce document doit être lisible et complet. ",
                                code: 'ATTH',
                                category: 'LOGACT',
                                label: "Attestation d'hébergement ou de domiciliation"}]
        }, {
          actual_housing_detail: %w[SD SAHFB CAMCAR HCAM HHOT HH HABMOB HHM LOGT],
          required_documents: [{ihm_order: "CXX",
                                type: "Attestation d’un travailleur social, d’une association ou certificat de domiciliation",
                                message: "Veuillez déposer ici une copie de l'attestation de domiciliation du travailleur social/association suivant votre dossier datant de moins de 3 mois.
Ce document doit être lisible et complet."
                               }]
        }, {
          actual_housing_detail: %w[HLM LLS LP LSLLP],
          required_documents: [{ihm_order: "C01",
                                type: "Contrat de bail",
                                message: "Veuillez déposer ici une copie de votre contrat de bail. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour'},
                               {ihm_order: "C01",
                                type: "Quittance de loyer ou attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués",
                                message: "Veuillez déposer ici une copie de votre dernière quittance de loyer. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',}
=begin
                               {ihm_order: "C01",
                                type: "Quittance de loyer",
                                message: "Veuillez déposer ici une copie de votre dernière quittance de loyer. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',
                                or: ["Attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués"]},

                               {ihm_order: "C01",
                                type: "Attestation du bailleur indiquant que vous êtes à jour de vos loyers et charges, ou tout moyen de preuve des paiements effectués",
                                message: "Veuillez déposer ici une attestation de votre bailleur datant de moins de 3 mois indiquant que votre situation locative est à jour. 
Attention, votre nom doit impérativement figurer sur ce document. 
Ce document doit être lisible et complet.",
                                code: 'JUSTL',
                                category: 'LOGACT',
                                label: 'Quittance ou tout autre justificatif de loyer à jour',
                                or: ["quittance de loyer"]}
=end
          ]
          
        }
      
      ]
      
      rules.each do |rule|
        next unless housing_request.housing_type_list.is_a?(Array) && (rule[:actual_housing_detail] & housing_request.housing_type_list).length > 0
        
        rule[:required_documents].each do |req_doc|
          find_or_create_required_document(req_doc[:ihm_order],
                                           "Logement Actuel",
                                           req_doc[:type],
                                           req_doc[:message],
                                           applicant.id.to_s,
                                           false,
                                           false,
                                           true,
                                           'SPR',
                                           req_doc[:code],
                                           req_doc[:category],
                                           false,
                                           req_doc[:label],
                                           nil,
                                           req_doc[:or])
        end
      end
      
      if is_main_applicant && housing_request.housing_type_list && housing_request.housing_type_list.include?('SLHT')
        #todo => sous locataire=>cas absent
        find_or_create_required_document(
          "CXX",
          "Logement Actuel",
          "(Vous êtes sous locataire)
(« Bail et dernière quittance de loyer ou attestation du bailleur indiquant que le locataire est à jour des loyers et charges ou tout moyen de preuve des paiements effectués. »

(Vous êtes hébergé dans un logement à titre temporaire)
« Attestation de la personne qui héberge »
", "Si vous êtes sous locataire, veuillez déposer ici une copie du bail et la dernière quittance de loyer ou une attestation du bailleur indiquant que le locataire est à jour des loyers et charges ou tout moyen de preuve des paiements effectués. Le document déposé doit être lisible et complet.

Si vous êtes hébergé dans un logement à titre temporaire, veuillez déposer ici une attestation de la personne qui héberge. Ce document doit être lisible et complet. 
",
          applicant.id.to_s,
          false,
          false,
          true,
          'SPL')
      end
    end
    
    def compute_motivation_documents(applicant_id, _is_main_applicant, is_untensed_area)
      all_motivations =  housing_request.motivations
      rules = [
        {
          motivation: '08',
          required_document: {ihm_order: "DXX",
                              type: "Certificat médical",
                              message: "Veuillez déposer ici une copie du certificat médical établi à votre nom. Ce document doit être lisible et complet. ",
                              code: 'JUSBA',
                              category: 'COMPH',
                              label: "Justificatif des besoins d'adaptation (certificat médical ou autre document)"}
        }, {
          motivation: '28',
          #todo: absent
          required_document: {ihm_order: "DXX",
                              type: 'Attestation de l\'employeur actuel ou futur justifiant votre mobilité professionnelle',
                              message: "Veuillez déposer ici une attestation de l'employeur actuel ou futur justifiant votre mobilité professionnelle. Ce document doit être lisible et complet. "}
        }, {
          motivation: '15',
          required_document: {ihm_order: "DXX", type: "Attestation de dépôt de demande de regroupement familial",
                              message: "Veuillez déposer ici une copie de l'attestation de dépôt de demande de regroupement familial. 
Ce document doit être lisible et complet. "}
        }, {
          motivation: '02',
          #todo: absent
          required_document: {ihm_order: "DXX", type: 'Toute pièce établissant de la démolition de votre logement actuel',
                              message: 'Veuillez déposer ici toute pièce établissant de la démolition de votre logement actuel. Ce document doit être lisible et complet. '}
        }, {
          motivation: '07',
          required_document: {ihm_order: "DXX", type: "Carte d’invalidité ou décision d’une commission administrative compétente (commissions des droits et de l’autonomie des personnes handicapées) ou d’un organisme de sécurité sociale",
                              message: "Veuillez déposer ici une copie de 
votre carte d'invalidité 
OU de la décision d’une commission administrative compétente (Commissions des droits et de l'autonomie des personnes handicapées) 
OU d’un organisme de sécurité sociale. Attention, le document doit impérativement faire figurer vos nom et prénom.
Ce document doit être lisible et complet. ",
                              code: 'CARINV',
                              category: 'COMPH',
                              label: "Carte d'invalidité ou décision de commission administrative compétente (MDPH...)"}
        }, {
          motivation: '05',
          required_document: {ihm_order: "D04", type: "Commandement de payer ou assignation à comparaître ou jugement prononçant l’expulsion ou commandement de quitter les lieux",
                              message: "Veuillez déposer ici une copie du commandement de payer OU de l'assignation à comparaître OU du jugement prononçant l'expulsion ou demande de quitter les lieux. 
Attention, le document doit impérativement faire figurer votre nom, prénom, adresse. 
Ce document doit être lisible et complet. ",
                              code: 'JUGEXP',
                              category: 'MOTF',
                              label: "Jugement d'expulsion"}
        }, {
          motivation: '22',
          #todo=> absent
          required_document: {ihm_order: "DXX",
                              type: 'Plan d\'apurement de la dette ou démarches en cours attestées par un travailleur social ou une association ou autre document démontrant les difficultés',
                              message: "Veuillez déposer ici une copie du plan d'apurement de la dette ou des démarches en cours attestées par un travailleur social ou une association ou autre document démontrant les difficultés. Ce document doit être lisible et complet. ",
                              code: 'LETI',
                              category: 'MOTIF',
                              label: "Lettre de l'intéressé ou d'un intervenant"}
        }, {
          motivation: '26',
          required_document: {ihm_order: "DXX",
                              type: "Agrément assistant(e) maternel(le)",
                              message: "Veuillez déposer ici une copie de votre agrément de profession. Attention, ce document doit impérativement faire figurer vos nom, prénom et adresse. Ce document doit être lisible et complet. "}
        }, {
          motivation: '06',
          required_document: {ihm_order: "DXX",
                              type: "Décision du juge prise en application de l'article 257 du code civil attestant de la situation d'urgence ou ordonnance de protection délivrée par le juge aux affaires familiales ou récépissé de dépôt de plainte",
                              message: "Veuillez déposer ici une copie de la décision du juge prise en application de l'article 257 du code civil attestant de la situation d'urgence ou ordonnance de protection délivrée par le juge aux affaires familiales ou récépissé de dépôt de plainte. Ce document doit être lisible et complet."
          }
        }, {
          motivation: '04',
          required_document: {ihm_order: "DXX",
                              type: "Lettre de congé du propriétaire ou jugement prononçant la résiliation du bail",
                              message: "Veuillez déposer ici une copie de la lettre de congé du propriétaire du logement OU le jugement prononçant la résiliation du bail de votre logement actuel. Ce document doit être lisible et complet.",
                              code: 'CONPR',
                              category: 'MOTIF',
                              label: 'Lettre de congé du propriétaire'}
        }, {
          motivation: '24',
          #todo=> absent
          required_document: {ihm_order: "D02",
                              type: "Document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement ou copie du jugement d'un tribunal statuant sur l'indécence du logement ou attestation de la CAF/MSA ou autre document démontrant l'indécence du logement",
                              message: "Veuillez déposer ici tout document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement ou copie du jugement d'un tribunal statuant sur l'indécence du logement ou attestation de la CAF/MSA ou autre document démontrant l'indécence du logement"}
        }, {
          motivation: '25',
          required_document: {
            ihm_order: "D03",
            type: "Logement insalubre ou dangereux :
Document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement, photos, jugement du tribunal, attestation de la CAF ou de la CMSA, copie d'une décision de l'administration (mise en demeure, arrêté préfectoral ou municipal d'insalubrité remédiable ou irrémédiable ou de péril, ou de fermeture administrative s'il s'agit d'un hôtel meublé) ou autre document démontrant l'insalubrité ou la dangerosité du logement ou de l'immeuble
Logement impropre à l'habitation :
Document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement, photos, jugement du tribunal, arrêté préfectoral de mise en demeure de faire cesser l'occupation des lieux ou autre document démontrant le caractère impropre à l'habitation",
            message: "En cas de logement insalubre ou dangereux, veuillez déposer ici tout document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement ou jugement du tribunal ou attestation de la CAF ou de la CMSA ou une copie d'une décision de l'administration (mise en demeure, arrêté préfectoral ou municipal d'insalubrité remédiable ou irrémédiable ou de péril, ou de fermeture administrative s'il s'agit d'un hôtel meublé) ou tout autre document démontrant l'insalubrité ou la dangerosité du logement ou de l'immeuble. Le document déposé doit être lisible et complet.

En cas de logement impropre à l'habitation, veuillez déposer ici un document établi par un service public, un travailleur social, un professionnel du bâtiment ou une association ayant pour objet l'insertion ou le logement, photos, jugement du tribunal, arrêté préfectoral de mise en demeure de faire cesser l'occupation des lieux ou autre document démontrant le caractère impropre à l'habitation. 
Le document déposé doit être lisible et complet."}
        }
      
      ]
      
      rules.each do |rule|
        if all_motivations && all_motivations.length >0 && all_motivations.include?(rule[:motivation])
          find_or_create_required_document(rule[:required_document][:ihm_order], "Motifs de la demande",
                                           rule[:required_document][:type],
                                           rule[:required_document][:message],
                                           applicant_id,
                                           false,
                                           false,
                                           true,
                                           'MD',
                                           rule[:required_document][:code],
                                           rule[:required_document][:category],
                                           false,
                                           rule[:required_document][:label])
        end
      end
    end
    
    def compute_dependents_documents(applicant_id, is_main_applicant, is_untensed_area)
      if is_main_applicant
        if housing_request.unborn_children&.length && housing_request.unborn_children[0] && housing_request.unborn_children[0][:anticipated_birth] &&
          Date.parse(housing_request.unborn_children[0][:anticipated_birth]) - 196.days < Date.today &&
          Date.today < Date.parse(housing_request.unborn_children[0][:anticipated_birth])
          find_or_create_required_document(
            "A06",
            "Dépendants",
            'Certificat de grossesse attestant que la grossesse est supérieure à 12 semaines',
            "Veuillez déposer ici une copie du certificat de grossesse datant de moins de 6 mois. Cette attestation peut vous être délivrée par votre médecin traitant. Ce document doit être lisible et complet.",
            applicant_id.to_s,
            false,
            false,
            true,
            'PC',
            'CGCH',
            'PCH',
            false,
            'Certificat de grossesse'
          )
        end
        
        if housing_request&.dependents && housing_request&.dependents&.length > 0
          housing_request.dependents.each do |dep|
            dep.to_h.symbolize_keys!
            dep_infos = dep.to_json
            if (dep[:rel_type] == "E" || dep[:coparentality] == "G" || dep[:coparentality] == "D")
              if (dep[:birth_date] && (Date.parse(dep[:birth_date]) <= Date.today - 18.years)) #todo
                find_or_create_required_document(
                  "AXX",
                  "Dépendants",
                  dep_infos + "####Carte nationale d'identité ou passeport ou titre de séjour ainsi que le récépissé de renouvellement de titre de séjour si celui-ci est expiré ",
                  "Carte nationale d'identité ou passeport ou titre de séjour ainsi que le récépissé de renouvellement de titre de séjour si celui-ci est expiré",
                  applicant_id.to_s,
                  false,
                  false,
                  true,
                  'PC',
                  'CICH',
                  'PCH',
                  true,
                  "Carte d'identité ou passeport d'une personne à charge"
                )
              end
              find_or_create_required_document(
                "A05",
                "Dépendants",
                dep_infos + '####Livret de famille ou acte de naissance',
                "Veuillez déposer ici une copie de votre livret de famille ou acte de naissance de tous les enfants devant vivre au foyer mentionnant votre nom. Ce document doit être lisible et complet.",
                applicant_id.to_s,
                false,
                false,
                true,
                'PC',
                'LFCH',
                'PCH',
                true,
                "Livret de famille ou acte d'État civil"
              )
            end
            next unless dep[:rel_type] == 'P' || dep[:rel_type] == 'A'
            
            find_or_create_required_document(
              "AXX",
              "Dépendants",
              dep_infos + "####Carte nationale d'identité ou passeport ou titre de séjour ainsi que le récépissé de renouvellement de titre de séjour si celui-ci est expiré ",
              "Carte nationale d'identité ou passeport ou titre de séjour ainsi que le récépissé de renouvellement de titre de séjour si celui-ci est expiré ",
              applicant_id.to_s,
              false,
              false,
              true,
              'PC',
              'CICH',
              'PCH',
              true,
              "Carte d'identité ou passeport d'une personne à charge"
            )
          end
        end
      end
    end
    
    def compute_disabilities_documents(applicant_id, is_main_applicant, is_untensed_area)
      if housing_request&.disability_details && housing_request&.disability_details.is_a?(Array) && housing_request&.disability_details&.length > 0
        housing_request.disability_details.each_with_index do |dis, i|
          find_or_create_required_document(
            "DXX",
            "handicap",
            '{"label":"Pour la personne handicapée n°' + (i + 1).to_s + '"}' + "****Carte d’invalidité ou décision d’une commission administrative compétente (commissions des droits et de l’autonomie des personnes handicapées) ou d’un organisme de sécurité sociale",
            '{"label":"Pour la personne handicapée n°' + (i + 1).to_s + '"}' + "****Veuillez déposer ici une copie de votre carte d'invalidité OU de la décision d’une commission administrative compétente (Commissions des droits et de l'autonomie des personnes handicapées) OU d’un organisme de sécurité sociale. Attention, le document doit impérativement faire figurer vos nom et prénom.Ce document doit être lisible et complet. 
",
            applicant_id.to_s,
            false,
            false,
            true,
            'HANDI',
            "CARINV",
            "COMPH",
            nil,
            nil
          )
        end
      end
    end
    
    # methode permettant d'associer au housing_request_folder un requiredDocument
    
    def find_or_create_required_document(ihm_order, nature, type, label, applicant_id, is_required_for_nud, is_required_for_consultation, is_required_for_application, category, sne_code = nil, sne_category = nil, same_for_all_applicant = false, sne_piece_label = nil, kyc_needed = false, type_ihm = nil, kyc_process = nil, or_required_document_types = [])
    #le champs type peut etre passé comme un array si la piece a eu plusieurs renommage de son type
      type_condition = type
      if (type.is_a? Array)
        type_condition =  {'$in' => type}
      end
     
      required_document = if same_for_all_applicant
                            Hermes::RequiredDocument.find_by(
                              type: type_condition,
                              applicant_id: {'$in' => all_applicant_ids}
                            )
                          else
                            Hermes::RequiredDocument.find_by(
                              type: type_condition,
                              applicant_id: applicant_id.to_s
                            )
                          end
      if required_document
        #si le nom courant de la piece a changé on la rennome a la sauvegarde
        if (type.is_a? Array) && required_document.type != type[0]
          required_document.type = type[0]
        end
        if is_required_for_nud
          unless required_document.folder_where_is_required_for_nud
            required_document.folder_where_is_required_for_nud = [id]
          end
          unless required_document.folder_where_is_required_for_nud.include?(id)
            required_document.folder_where_is_required_for_nud.push(id)
          end
        else
          required_document.folder_where_is_required_for_nud&.delete(id)
        end
        if is_required_for_consultation
          unless required_document.folder_where_is_required_for_consultation
            required_document.folder_where_is_required_for_consultation = [id]
          end
          unless required_document.folder_where_is_required_for_consultation.include?(id)
            required_document.folder_where_is_required_for_consultation.push(id)
          end
        else
          
          required_document.folder_where_is_required_for_consultation&.delete(id)
        end
        if is_required_for_application
          unless required_document.folder_where_is_required_for_application
            required_document.folder_where_is_required_for_application = [id]
          end
          unless required_document.folder_where_is_required_for_application.include?(id)
            required_document.folder_where_is_required_for_application.push(id)
          end
        else
          required_document.folder_where_is_required_for_application&.delete(id)
        end
        required_document.save # TODO: OPTIMISATION ne pas save si pas de changement
      end
      required_document_ids.push(
        if !required_document
          Hermes::RequiredDocument.create(
            ihm_order: ihm_order,
            nature: nature,
            type: (type.is_a? Array)? type[0] : type,
            label: label,
            applicant_id: applicant_id.to_s,
            status: 'Pièce manquante',
            folder_where_is_required_for_nud: is_required_for_nud ? [id] : [],
            folder_where_is_required_for_consultation: is_required_for_consultation ? [id] : [],
            folder_where_is_required_for_application: is_required_for_application ? [id] : [],
            sne_code: sne_code,
            sne_category: sne_category,
            sne_piece_label: sne_piece_label,
            category: category,
            user_id: user_id,
            type_ihm: type_ihm,
            or_required_document_types: or_required_document_types
          ).id
        else
          required_document.id
        end
      )
    end
    
    def status_for_required_document_list(required_document_list)
      required_document_list.any? { |rd| !rd.declared } ? 'Dossier incomplet'  : 'Dossier validé' # only use declared for status devops 13017
    end
   
    def compute_statuses
      required_documents_for_nud = []
      required_documents_for_consultation = []
      required_documents_for_application = []

      required_documents.each do |required_document|
        required_document.reload
        if required_document.folder_where_is_required_for_nud.include? id
          required_documents_for_nud.push(required_document)
        end
        if required_document.folder_where_is_required_for_consultation.include? id
          required_documents_for_consultation.push(required_document)
        end
        if required_document.folder_where_is_required_for_application.include? id
          required_documents_for_application.push(required_document)
        end
      end
      
      self.nud_status = status_for_required_document_list(required_documents_for_nud) #+ "pour consultation."
      self.consultation_status = status_for_required_document_list(required_documents_for_consultation) #+ "pour consultation."
      self.application_status = status_for_required_document_list(required_documents_for_application) #+ "pour candidater."
    
    end
    
    def before_save_action
      compute_statuses
    end
    
    def before_create_action
      self.user_id = housing_request&.user_id
    end
    
    def after_update_action
      if application_status_changed? || consultation_status_changed? || nud_status_changed?
        housing_request.reload.update_request_status 
        housing_request.save if housing_request.request_status_changed? || housing_request.form_status_changed?
      end
    end
    
    def notify_applicant
      # Temporary_disabled_devops_7390
      # if application_status_changed?
      #   case application_status
      #     when 'Erreurs de validation dans les pieces fournies' # disabled
      #       required_docs = required_documents.select{ |rd| rd.folder_where_is_required_for_application.include?(id) && rd.invalid_doc? }
      #       housing_request.notify(
      #         key: NOTIFICATION_KEYS[:housing_request][:documents_application_ko],
      #         main_only: true,
      #         datas: {
      #           html_doc_list: "<ul>#{required_docs.map{ |doc| "<li>#{doc.nature} : #{doc.label}</li>" }.join}</ul>",
      #         },
      #         channels: ['email']
      #       )
      #     when 'Dossier incomplet' # disabled
      #       required_docs = required_documents.select{ |rd| rd.folder_where_is_required_for_application.include?(id) && rd.missing? }
      #       housing_request.notify(
      #         key: NOTIFICATION_KEYS[:housing_request][:documents_application_missing],
      #         main_only: true,
      #         datas: {
      #           html_doc_list: "<ul>#{required_docs.map{ |doc| "<li>#{doc.nature} : #{doc.label}</li>" }.join}</ul>",
      #         }, 
      #         channels: ['email']
      #       )
      #     else
      #     nil
      #   end
      # end
      # if consultation_status_changed?
      #   case consultation_status
      #   when 'Erreurs de validation dans les pieces fournies'
      #     required_docs = required_documents.select {|rd| rd.folder_where_is_required_for_consultation.include?(id) && rd.invalid_doc?}
      #     housing_request.notify(
      #       key: NOTIFICATION_KEYS[:housing_request][:documents_consultation_ko],
      #       main_only: true,
      #       datas: {
      #         html_doc_list: "<ul>#{required_docs.map {|doc| "<li>#{doc.nature} : #{doc.label}</li>"}.join}</ul>",
      #       },
      #       channels: ['email']
      #     )
      #   else
      #     nil
      #   end
      # end
    end
  end
end