# frozen_string_literal: true

module Hermes
  module BoClient
    class BaseTask
      ParamDefinition = Struct.new(:name, :type, :options)
      
      DEFAULT_ENDPOINT = '/ikos/ws/rest/PMA/IntegrationPatrimoine/v1/'
      
      ### Class attributes accessors
      
      class << self
        def endpoint_value
          @endpoint ||= DEFAULT_ENDPOINT
        end
        
        def mapping_value
          raise 'Set your mapping using BaseTask.task_name' unless @mapping
          
          @mapping
        end
        
        def type_flux_value
          raise 'Set your type_flux using BaseTask.type_flux' unless @type_flux
          
          @type_flux
        end
        
        def param_definitions
          @param_definitions ||= {}
        end
        
        ### TASK DSL
        
        # If there is no task name, set it to `:none`
        attr_writer :mapping, :task_name, :type_flux
        
        def mapping value
          @mapping = value
        end
        
        def task_name  value
        @task_name = value 
        end
        
        def method_name (value ="post")
          @method_name = value
        end
        
        def type_flux value
          @type_flux = value
        end
        
        def with_process_instance_id?
          @with_process_instance_id ||= true
        end
        
        def no_process_instance_id
          @with_process_instance_id = false
        end
        
        # block will be executed in `instance_eval`
        def endpoint(value = nil, &block)
          @endpoint = value || block
        end
        
        def param(name, type = nil, options = {})
          param_definitions[name.to_sym] = ParamDefinition.new(name, type, options)
          define_method(name.to_s.underscore) {get_parameter(name.to_sym)}
          define_method("#{name.to_s.underscore}=") {|value| set_parameter(name.to_sym, value)}
        end
        
        def params(*parameter_names)
          parameter_names.each {|parameter_name| param(parameter_name)}
        end
      end
      ### Base Task
      
      attr_accessor :workflow, :options
      
      def parameters
        @parameters ||= {}
      end
      
      def initialize(options = {})
        @options = options
        parameters.each {|k, v| send("#{k.to_s.underscore}=", v)}
      end
      
      def task_name
        self.class.task_name_value
      end
      
      def type_flux
        self.class.type_flux_value
      end
      
      def mapping
        self.class.mapping_value.constantize
          .new(@options)
      end
      
      def endpoint
        val = self.class.endpoint_value
        Rails.logger.debug "Class #{val.class}"
        return instance_eval(&val) if val.is_a? Proc
        
        val
      end
      
      def method
        val = self.class.method_name
        Rails.logger.debug "Class #{val.class}"
        return instance_eval(&val) if val.is_a? Proc
        
        val
      end
      
      def get_parameter(key)
        parameters[key.to_sym]
      end
      
      def set_parameter(key, value)
        parameters[key] = value
      end
      
      def execute(client = Hermes::BoClient::Client.new)
        #return {exectuted: true, xml: to_xml} if Rails.env == 'test'
        
        ExecuteSopraTaskJob.perform_later(endpoint, validate_xml(to_xml), method, mapping.main_object.class.name, mapping.main_object.id.to_s)
      end
      
      def to_xml
        namespaces = {
          'xmlns:ns2' => 'http://v1.als.ikos.webservice.immobilier.sopra/patrimoineALS'
        }
        builder = Nokogiri::XML::Builder.new(encoding: 'UTF-8') do |b|
          b[:ns2].patrimoine(namespaces) do
            b.informations_flux do
              b.parent.namespace = nil
              b.version_flux 01
              b.send('type_flux', type_flux) unless type_flux == :none
              b.date_flux Date.today
            end
            b.bien do
              b.parent.namespace = nil
              # self.parameters.each do |k,v|
              #   b.send(k, v)
              # end
              self.class.param_definitions.each do |key, struct|

                if(struct[:name]== :orientations_bailleur)
                  if (mapping.send("orientations_bailleur") && mapping.send("orientations_bailleur").length)
                    b.orientations_bailleur do
                      mapping.send("orientations_bailleur").each do |orientation|
                        b.orientation orientation
                      end
                    end
                  end
                else
                  seeked_elt = mapping.send(struct[:name])
                  next unless seeked_elt
                
                  b.send(key, seeked_elt)
                end
              
              end
              # self.mapping.each do |k, v|
              #   b.send(v, @object.send(k))
              # end
            end
          end
          
          # b.attribute('workflowNumber', self.workflow.id)
          # b.attribute('processInstanceId', self.workflow.process_instance_id) if self.class.with_process_instance_id?
        end
        
        builder.to_xml
      end
      
      def validate_xml xml = nil
        xml ||= to_xml
        schema_path = [Hermes::Engine.root, %w[spec libs bo_client tasks], @xsd].flatten.join('/').to_s
        xsd = Nokogiri::XML::Schema(File.read(schema_path))
        xml_validation_report = xsd.validate(Nokogiri::XML(xml))
        mapping.main_object.set(updated_at: DateTime.current, last_xml_validation_report: xml_validation_report.to_s)
        
        return xml
      end
      
      private
      
      def parameter_to_xml(builder, key, value)
        param_def = self.class.param_definitions[key]
        
        value ||= param_def.options[:default]
        
        if param_def.options[:in]
          raise InvalidValue unless param_def.options[:in].include?(value)
        end
        
        case param_def.type.to_s
        when 'Array'
          builder.parameter(key, value.join(';'), {value_tag: param_def.options[:value_tag]})
        when 'Hash'
          hkeys = []
          hvalues = []
          value.each do |hk, hv|
            hkeys << hk
            hvalues << hv
          end
          builder.parameter(param_def.options[:keys_name], hkeys.join(';'))
          builder.parameter(param_def.options[:values_name], hvalues.join(';'))
        when 'Integer'
          builder.parameter(key, value.to_i, {value_tag: param_def.options[:value_tag]})
        when 'Boolean'
          builder.parameter(key, true, {value_tag: param_def.options[:value_tag]}) do |xml|
            xml.boolean value.to_s
          end
        else
          builder.parameter(key, value.to_s, {value_tag: param_def.options[:value_tag]})
        end
      end
    end
  end
end
