require 'rails_helper'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  include Rack::Test::Methods
  
  describe Api::Dmo::NotificationCenter::NotificationsController, :type => :api do
    
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
    let!(:user){ create(:user) }
    let!(:applicant){ create(:applicant, notifications_count: 5) }

    describe 'Actions' do
      before :each do
        authenticated_header(request, user)
      end

      it 'return notifications' do
        get :index, targetable_type: 'Hermes::Applicant', id: applicant.id

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(5)
      end

      it 'return notification' do
        notif = applicant.notifications.first
        get :show, targetable_type: 'Hermes::Applicant', id: applicant.id, notification_id: notif.id

        expect(response).to have_http_status(200)
        expect(json['data']['id']).to eq(notif.id.to_s)
      end

      it 'delete notification' do
        notif = applicant.notifications.first

        expect { delete :destroy, targetable_type: 'Hermes::Applicant', id: applicant.id, notification_id: notif.id }.to change(applicant.notifications, :count).by(-1)
        expect(response).to have_http_status(204)
      end

      it 'open notification' do
        notif = applicant.notifications.first
        put :open, targetable_type: 'Hermes::Applicant', id: applicant.id, notification_id: notif.id

        expect(response).to have_http_status(200)
        expect(json['data']['attributes']['opened_at']).not_to be_nil
      end

      it 'move notification' do
        notif = applicant.notifications.first
        get :move, targetable_type: 'Hermes::Applicant', id: applicant.id, notification_id: notif.id

        expect(response).to have_http_status(200)
        expect(json['links']['notifiable_url']).to eq(notif.notifiable.notifiable_path)
      end

      it 'open all notifications' do
        post :open_all, targetable_type: 'Hermes::Applicant', id: applicant.id

        expect(response).to have_http_status(200)
        expect(applicant.has_unopened_notifications?).to be(false)
      end

      it 'open last n notifications with limit filter' do
        post :open_all, targetable_type: 'Hermes::Applicant', id: applicant.id, limit: 3

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(3)
        expect(applicant.unopened_notifications.count).to eq(2)
      end

      it 'use filter params' do
        expect(applicant.notifications.count).to eq(5)

        get :index, targetable_type: 'Hermes::Applicant', id: applicant.id, filter: :opened

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(0)

        applicant.notifications.last.open!
        get :index, targetable_type: 'Hermes::Applicant', id: applicant.id, filter: :unopened

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(4)

        get :index, targetable_type: 'Hermes::Applicant', id: applicant.id, filter: :all

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(5)
      end

      it 'use limit params' do
        expect(applicant.notifications.count).to eq(5)

        get :index, targetable_type: 'Hermes::Applicant', id: applicant.id, limit: 2

        expect(response).to have_http_status(200)
        expect(json['data'].count).to eq(2)
      end
    end
  end

  # Swagger
  require_relative '../../../swagger_helper'

  describe Hermes::Api::Dmo::NotificationCenter::NotificationsController, type: :request, swagger_doc: 'api/swagger/hermes/swagger_hermes.json' do
    TAGS_API_DMO_NOTIFICATIONS = 'API_DMO_NOTIFICATIONS'
    
    include JwtAuthHelper
    extend HermesSwaggerJsonParams
    include SwaggerJsonAttributes
    let!(:admin) {create(:administrator)}
    
    context 'NotificationCenter' do
      path '/api/dmo/applicants/{applicant_id}/notifications' do
        get 'Retrieves all notifications' do
          tags TAGS_API_DMO_NOTIFICATIONS
          produces 'application/json'
          
          parameter name: :applicant_id, in: :path, type: :string,  required: true
          parameter name: :filter, in: :query, type: :string, enum: [:all, :opened, :unopened], required: false
          parameter name: :limit, in: :query, type: :integer, required: false, description: 'Limit selection (order: created_at DESC)'
          security [Bearer: {}]
          paginator
          
          response '200', 'notifications found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:notifications) { create_list(:notification, 5, targetable: applicant ) }
            let!(:applicant_id) { applicant.id }
            
            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) { notifications }
            end
          end
        end
      end

      path '/api/dmo/applicants/{applicant_id}/notifications/{notification_id}' do
        get 'Retrieves a notification' do
          tags TAGS_API_DMO_NOTIFICATIONS
          produces 'application/json'
          
          parameter name: :applicant_id, in: :path, type: :string,  required: true
          parameter name: :notification_id, in: :path, type: :string,  required: true
          security [Bearer: {}]
          
          response '200', 'notification found' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:notification) { create(:notification, targetable: applicant ) }
            let!(:notification_id) { notification.id }
            let!(:applicant_id) { applicant.id }
            
            include_context 'with integration test'
            include_context 'valid jsonapi'            
          end
        end

        delete 'Delete a notification' do
          tags TAGS_API_DMO_NOTIFICATIONS
          produces 'application/json'
          
          parameter name: :applicant_id, in: :path, type: :string,  required: true
          parameter name: :notification_id, in: :path, type: :string,  required: true
          security [Bearer: {}]
          
          response '204', 'notification deleted' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:notification) { create(:notification, targetable: applicant) }
            let!(:notification_id) { notification.id }
            let!(:applicant_id) { applicant.id }
            
            run_test!
          end
        end
      end

      path '/api/dmo/applicants/{applicant_id}/notifications/{notification_id}/open' do
        put 'Open a notification' do
          tags TAGS_API_DMO_NOTIFICATIONS
          produces 'application/json'
          
          parameter name: :applicant_id, in: :path, type: :string,  required: true
          parameter name: :notification_id, in: :path, type: :string,  required: true
          security [Bearer: {}]
          
          response '200', 'notification opened' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:notification) { create(:notification, targetable: applicant ) }
            let!(:notification_id) { notification.id }
            let!(:applicant_id) { applicant.id }
            
            include_context 'with integration test'
            include_context 'valid jsonapi'
          end
        end
      end

      path '/api/dmo/applicants/{applicant_id}/notifications/open_all' do
        post 'Open all notifications' do
          tags TAGS_API_DMO_NOTIFICATIONS
          produces 'application/json'
          
          parameter name: :applicant_id, in: :path, type: :string,  required: true
          parameter name: :limit, in: :body, type: :integer, required: false, description: 'Limit selection (order: created_at DESC)'
          security [Bearer: {}]
          
          response '200', 'notifications opened' do
            let(:Authorization) { "Bearer #{token_for(admin)}" }
            let!(:applicant) { create(:applicant) }
            let!(:notification) { create(:notification, targetable: applicant ) }
            let!(:notification_id) { notification.id }
            let!(:applicant_id) { applicant.id }
            
            include_context 'with integration test'
            include_context 'valid jsonapi collection' do
              let(:expected_collection) { applicant.notifications }
            end
          end
        end
      end
    end
  end
end