# frozen_string_literal: true

require 'rails_helper'

# %w[
#   applicant
#   housing_request
# ].each { |factory_name|require Hermes::Engine.root.join('spec', 'factories', "#{factory_name.pluralize}_factory").to_s }

RSpec.describe Hermes::HousingRequest, type: :model do
  let!(:user) {create(:user)}
  describe 'basic controls' do
    let!(:user1) {create(:user)}
    let!(:applicant1) do
      create(
        :applicant,
        user: user1,
        monthly_charges: {}
      )
    end
    
    it 'has a valid factory' do
      expect(FactoryBot.build(:housing_request, main_applicant: applicant1)).to be_valid
    end
  end
  
  context 'with simple setup' do
    before :each do
      $redis.flushall
    end
    let!(:user1) {create(:user)}
    let!(:user2) {create(:user)}
    let!(:user3) {create(:user)}
    let!(:applicant1) {
      create(
        :applicant,
        user: user1,
        monthly_charges: {},
      )
    }
    let!(:applicant2) {
      create(
        :applicant,
        user: user2
      )
    }
    let!(:applicant3) {
      create(
        :applicant,
        user: user3
      )
    }
    
    let!(:housing_request1) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        seeked_departments: []
      )
    end
    
    let!(:housing_request2) do
      create(
        :housing_request,
        is_active: true,
        main_applicant: applicant1,
        coapplicants: [
          {rel_type: 'Conjoint (e)', id: applicant2.id.to_s},
          {rel_type: 'Parent (e)', id: applicant3.id.to_s}
        ]
      )
    end
    
    describe 'Eligibility query are valid' do
      it 'should include right typology ranges' do
        typology_range1 = housing_request1.update_typology_range
        typology_range2 = housing_request2.update_typology_range
        range1_to_i = typology_range1.map {|typology| Integer(typology.sub('T', ''))}
        range2_to_i = typology_range2.map {|typology| Integer(typology.sub('T', ''))}
        
        (1..2).each do |n|
          (1..2).each do |i|
            expect(binding.local_variable_get("range#{n}_to_i".to_sym)).to include i
          end
        end
        expect(range1_to_i.max).to be 2
        expect(range2_to_i).to include 3
        expect(range2_to_i.max).to be 3
      end
      
      it 'eligibility should incorpore typology' do
        housing_request1.update_typology_range
        typology_filter1 = housing_request1.typology_filter
        expect(typology_filter1[:typology][:$in]).to include 'T1'
        expect(typology_filter1[:typology][:$in]).to include 'T2'
      end
      
      describe 'edit action dont duplicate coapplicants' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.seeking_elevator = false
          housing_request2.save
          housing_request2.reload
          housing_request2.coapplicants = [{rel_type: 'M', id: applicant3.id.to_s}]
          housing_request2.save
          applicant3.destroy
          housing_request2.reload
          housing_request2.forcing_duplicate = true
          housing_request2.save
          housing_request2.reload
        end
        it 'coapplicants doesnt change' do
          expect(housing_request2.reload.applicant_ids.length).to eq(0)
        end
      end
    end
    
    describe "#update_seeked_location" do
      before :each do
      $redis.flushall
      end
      let(:list) do
        {'nom': 'Fréjus', 'code': '83061', 'codeDepartement': '81', 'codesPostaux': ['83900', '83370', '']}
      end
      let(:lists) do
        [
          {'nom': 'Paris', 'code': '75001', 'codeDepartement': '75', 'codesPostaux': ['75006', nil]},
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
          {'nom': 'Paris', 'code': '75002', 'codeDepartement': '75', 'codesPostaux': ['75006', '93000']}
        ]
      end
      let(:list_without_postal_codes) do
        [
          {'nom': 'Meaux', 'code': '77000', 'codeDepartement': '77'},
        ]
      end

      let(:broken_list) do
        {'codePostal': '77013', 'foo': 'bir'}
      end
      
      it 'sets seeked fields on save' do
        housing_request1.seeked_localisation_list = [list]
        expect {housing_request1.save}.to change {housing_request1.seeked_departments}.from([]).to(['81'])
                                            .and change {housing_request1.truly_seeked_departments}.from([]).to(['81', '83'])
                                                   .and change {housing_request1.seeked_district_postal_codes}.from([]).to(['83370', '83900'])
                                                          .and change {housing_request1.seeked_districts}.from([]).to(['Fréjus'])
      end
      
      it 'ensures the sorted uniqueness of seeked fields' do
        housing_request1.seeked_localisation_list = lists + [list]
        housing_request1.save
        expect(housing_request1.seeked_departments).to eq ['75', '77', '81']
        expect(housing_request1.truly_seeked_departments).to eq ['75', '77', '81', '83']
        expect(housing_request1.seeked_district_postal_codes).to eq ['75006', '83370', '83900', '93000']
        expect(housing_request1.seeked_districts).to eq ['Fréjus', 'Meaux', 'Paris']
      end
      
      it 'get truly_seeked_departments from insee code' do
        housing_request1.seeked_localisation_list = list_without_postal_codes
        expect {housing_request1.save}.to change {housing_request1.truly_seeked_departments}.from([]).to(['77'])
      end

      it 'returns error' do
        housing_request1.seeked_localisation_list = [broken_list]
        housing_request1.save
        expect(housing_request1.errors.full_messages.first).to match /non-conform seeked_localisation_list/
      end
    end
    
    describe 'Relationships are valid' do
      before :each do
        $redis.flushall
      end
      it 'should has correct housing_requests situation' do
        expect(applicant1.first_active_request.id).to eq(housing_request2.id)
        expect(applicant1.housing_requests.length).to eq(2)
        expect(applicant1.housing_requests.actives.length).to eq(1)
      end
      
      it 'should has correct values' do
        revenues_sum = (applicant1.monthly_revenues || 0) + (applicant2.monthly_revenues || 0)
        expect(housing_request2.monthly_revenues).to eq(revenues_sum)
      end
      
      it 'should has correct coapplicants' do
        coapplicants = housing_request2.coapplicants
        expect(coapplicants.length).to eq(2)
        expect(coapplicants.map {|coapplicant| coapplicant[:rel_type]}).to include('Conjoint (e)').and include('Parent (e)')
        expect(coapplicants.map {|coapplicant| coapplicant[:id]}).to include(applicant2.id).and include(applicant3.id)
      end
      
      describe 'after coapplicants update' do
        before :each do
          housing_request2.coapplicants = [
            {rel_type: 'Conjoint (e)', id: applicant3.id}
          ]
          housing_request2.save
        end
        
        it 'should has updated with valid data' do
          r_request = housing_request2.reload
          expect(r_request.applicant_ids.length).to be 1
          expect(r_request.applicant_ids.first).to eq applicant3.id
          expect(r_request.coapplicants.first[:rel_type]).to eq 'Conjoint (e)'
        end
      end
    end
    
    describe 'applicant should have only public offer as eligibility when last active request becomes inactive' do
      it 'should has first active request with valid data' do
        expect(housing_request2.main_applicant.housing_requests.actives.count).to eq 1
        expect(housing_request2.is_active).to be true
        expect(housing_request2.main_applicant.first_active_request.id.to_s).to eq housing_request2.id.to_s
      end
      
      context 'is_active is set to false' do
        before :each do
          housing_request2.update(is_active: false)
        end
        
        it 'should has is_active == false' do
          expect(housing_request2.reload.is_active).to be false
        end
        
        it 'should has its main_applicant.eligibility == {is_public: true}' do
          expect(housing_request2.reload.main_applicant.eligibility_query).to eq '{"is_public":true}'
        end
      end
    end
    
  end

  describe "2 active housing request with same nud is forbidden" do
    let!(:applicant) {create(:applicant)}
    let!(:applicant2) {create(:applicant)}
    let!(:request1) {create(:housing_request, main_applicant: applicant, is_active: true)}
    let!(:request2) {create(:housing_request, main_applicant: applicant2, is_active: false)}
    let!(:nud_test) {create(:nud, housing_request: request1, unique_number: "011752019061716361")}
    let!(:nud_test2) {create(:nud, housing_request: request2, unique_number: "011752019061716361")}
    it "should throw error" do
      expect {request2.is_active = true }.to change {request2.valid?}.from(true).to(false)
    end
  end

  describe "generate als id" do 
    let!(:applicant1) {create(:applicant)}
    let!(:applicant2) {create(:applicant)}
    let!(:hr1) { create(:housing_request, main_applicant: applicant1)}
    let!(:hr2) { create(:housing_request, main_applicant: applicant2)}

    it 'job generate 2 als_id' do
      expect( enqueued_jobs.select {|job| job[:job] == AlsIdUniqJob}.length).to eq(2)
      drain_all_jobs
      expect(hr1.reload.als_id).not_to eq(hr2.reload.als_id)
    end 

    it 'cannot save with same als_id' do 
      Hermes::HousingRequest.remove_indexes
      Hermes::HousingRequest.create_indexes
      AlsIdUniqJob.perform_now(hr1.id.to_s, "test")
      AlsIdUniqJob.perform_now(hr2.id.to_s, "test")
      expect(hr1.reload.als_id).to eq("test")
      expect(hr2.reload.als_id).to eq(nil)
    end 

    it 'not regenerate als_id ' do 
      drain_all_jobs
      als_id = '0000000001'
      expect(hr1.reload.als_id).to eq(als_id)
      10.times do |i|
        AlsIdUniqJob.perform_now(hr1.to_s)
      end
      expect(hr1.reload.als_id).to eq(als_id)
    end 

  end 

  describe "check spec via csv" do
    path = [Hermes::Engine.root, %w[spec xml_bouchon], "Nomenclature Cerfa v4 - Sheet1.csv"  ].flatten.join('/').to_s
    only_line= false
    cahier_tests_cerfa_v4 = CSV.read(path)
    cahier_tests_cerfa_v4.each_with_index do |line,i|
      if(line[1] && line[2] && line[3] && line[4] && i>0 && (!only_line || i +1 ==only_line))
        describe "test ligne "+ (i+1).to_s + " : " + line[0] + " ----> check if "+line[1]+" is well extracted in xml with unique number "+line[3] do
          let!(:applicant1) {create(:applicant,
            postal_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            owned_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            actual_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},

            )}
          let!(:housing_request) { create(:housing_request, main_applicant: applicant1,   
            seeked_localisation_list:[{code: "33318", nom: "valette"}],
            postal_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            owned_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},
            actual_housing_details: {"district"=> "pantin", "address_insee_code"=> "75056"},)}
          let!(:nud) {create(:nud, housing_request: housing_request,  unique_number: line[3])}
          before :each do
            housing_request.update_request_from_sne(user)
          end
          it ('should ave value eq to ' + line[4]) do
            path_in_model = line[1].split(".")
            collected_value=housing_request
            if(housing_request.xml)
              if(path_in_model[0] == "coapplicants")
                seached_applicant = Hermes::Applicant.find(housing_request.coapplicants[path_in_model[1].to_i][:id])
                if(path_in_model[2] == "rel_type")
                  collected_value = housing_request.coapplicants[path_in_model[1].to_i][:rel_type]
                else
                  collected_value = seached_applicant.send(path_in_model[2])
                end
              else
               
                path_in_model.each do |key|
                  if collected_value.respond_to?(key)
                    collected_value = collected_value.send(key)
                  elsif (collected_value[key.to_sym] rescue false)
                    collected_value = collected_value[key.to_sym]
                  elsif (collected_value[key.to_i] rescue false)
                    collected_value = collected_value[key.to_i]
                  else
                    collected_value = collected_value[key.to_s]
                  end
                end
              end
              value_to_compare = line[4]
              case line[4] # a_variable is the variable we want to compare
                when "TRUE"    #compare to 1
                  value_to_compare = true
                when "FALSE"    #compare to 2
                  value_to_compare = false
                when "NIL"    #compare to 2
                  value_to_compare = nil
              end
              if( collected_value.is_a?(Date))
                value_to_compare = Date.parse(line[4])
              end
              if( collected_value.is_a?(Integer))
                value_to_compare = line[4].to_i
              end
              expect(collected_value).to eq(value_to_compare)
            else
              ap "le fichier XML "+ line[3] + " est manquant et ne permet pas de réaliser le test n°"+ i.to_s
            end
          end
        end
      end
    end
  end

  describe "update_from_sne" do
    let!(:applicant1) {create(:applicant)}
    let!(:housing_request) { create(:housing_request, main_applicant: applicant1, user: user, consent_level:1)}
    let!(:nud) {create(:nud, housing_request: housing_request, department: nil,  unique_number: "1110121021518G1104")}

    describe "first test" do 
      before :each do
        housing_request.update_request_from_sne(user)
      end
      it 'should add coapplicants to request' do
        # on check qu'on a le bon nombre d'applicant
        expect(housing_request.coapplicants.length).to eq(8)
        expect(Hermes::Applicant.all.length).to eq(9)
        applicant_id = housing_request.coapplicants[0][:id]
        applicant = Hermes::Applicant.find(applicant_id)
        expect(applicant.nationality).to eq("1")
        #on check qu'ils sont bien updatés
        housing_request.main_applicant.main_employer_name = "minie"
        applicant.nationality = "2"
        applicant.main_employer_name = "picsou"
        applicant.save
        expect(applicant.nationality).to eq("2")

        housing_request.update_request_from_sne(user)
        #on check que les données sont updatées
        expect(applicant.reload.nationality).to eq("1")
        #on check que le main employer name est conservé
        expect(applicant.reload.main_employer_name).to eq("picsou")
        expect(housing_request.main_applicant.reload.main_employer_name).to eq("minie")
        expect(housing_request.consent_level).to eq(1)
        expect(applicant.reload.id.to_s).to eq(applicant_id)
        expect(Hermes::Applicant.all.length).to eq(9)
        applicant.first_name = "momo"
        applicant.last_name = "momo"
        applicant.save

        housing_request.reload.update_request_from_sne(user)
        expect(Hermes::Applicant.where(id: applicant.id).length).to eq(0)
        coloc = Hermes::Applicant.find(housing_request.coapplicants[5][:id])
        housing_request.coapplicants.each do |coapp|
          expect(Hermes::Applicant.find(coapp[:id]).user_id).to eq(user._id)
        end
        # Not working
        # coloc.unemployment_period = true
        # coloc.save
        # housing_request.reload.update_request_from_sne(user)
        # expect(coloc.reload.unemployment_period).to eq(true)
        nud.unique_number = "1110121021545G1104"
        nud.save
        housing_request.reload.update_request_from_sne(user)
        housing_request.coapplicants.each do |coapp|
          applic = Hermes::Applicant.find(coapp[:id])
          expect(applic.user_id).to eq(user._id)
        end
        nud.unique_number = "1110121021518G1104"
        nud.save
        housing_request.reload.update_request_from_sne(user)
        housing_request.coapplicants.each do |coapp|
          applic = Hermes::Applicant.find(coapp[:id])
          expect(applic.user_id).to eq(user._id)
        end

        housing_request.main_applicant.workplace_locality = "pantin"
        housing_request.main_applicant.save
        nud.unique_number = "0331120001997G1175"
        nud.save
        housing_request.reload.update_request_from_sne(user)
          expect(housing_request.reload.main_applicant.workplace_locality).to eq(nil)
       
      end
    end
  end

  describe "update_request_supplement" do 
    let!(:applicant1) {create(:applicant)}
    let!(:hr) { create(:housing_request, main_applicant: applicant1, user: user)}
    let!(:nud) {create(:nud, housing_request: hr, department: nil,  unique_number: "1110121021518G1104")}
    let!(:offer) {create(:housing_offer)}

    it 'not trigger when default provider' do 
      hr.update_request_from_sne(hr.user)
      expect(hr.reload.cotations).to eq(nil)
    end 

    it 'when IMHOWEB' do 
      nud.set(provider: 'IMHOWEB')
      hr.update_request_from_sne(hr.user)
      expect(hr.reload.cotations).to eq({"21003"=>136, "21231"=>136})
      expect(hr.reload.supplement_status).to eq("Active")
    end 

    it 'when applying' do 
      nud.set(provider: 'IMHOWEB')
      app = Hermes::HousingApplication.create(housing_request: hr, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      expect(hr.reload.supplement_can_apply).to eq(true)
      expect(app.valid?).to eq(true)
      expect(hr.reload.supplement_status).to eq("Active")
      expect(hr.reload.cotations).to eq({"21003"=>136, "21231"=>136})
      expect(hr.reload.cotations).to eq(app.cotations)
    end 

    it 'when applying when cannot' do 
      nud.set(provider: 'IMHOWEB', unique_number: "1110121021518G1105")
      hr.set(unique_numbers: ['1110121021518G1105'])
      app = Hermes::HousingApplication.create(housing_request: hr, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      # TODO : remove when imhoweb fixed
      expect(app.valid?).to eq(true)
      # expect(hr.reload.supplement_can_apply).to eq(false)
      # expect(app.valid?).to eq(false)
      # expect(hr.reload.cotations).to eq({"21003"=>136, "21231"=>136})
    end 

    it 'when no localisationsSouhaitees' do 
      nud.set(provider: 'IMHOWEB', unique_number: "1110121021518G1106")
      hr.set(unique_numbers: ['1110121021518G1106'])
      hr.update_request_supplement(hr.user)
      expect(hr.reload.cotations).to eq({})
    end 

    it 'when sopra error' do 
      nud.set(provider: 'IMHOWEB', unique_number: "not_found")
      hr.set(unique_numbers: ['not_found'])
      error = hr.update_request_supplement(hr.user)
      expect(hr.reload.cotations).to eq(nil)
      expect(error[:errors]).to eq('missing test file')
    end 

    it 'when applying with timeout' do 
      nud.set(provider: 'IMHOWEB')
      HERMES_CONFIG['sleeptimeout'] = 11
      app = Hermes::HousingApplication.create(housing_request: hr, housing_offer: offer, status: 'EN ATTENTE', visit_status: 'VISITE PROPOSEE')
      HERMES_CONFIG['sleeptimeout'] = nil

      expect(app.valid?).to eq(true)
      expect(hr.reload.cotations).to eq(nil)
    end 
    

  end 

  describe "After actions" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '12345678901234') }
    let!(:housing_offer) { create(:housing_offer, lessor: lessor) }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity.siret) }  
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant: applicant, request_status: 'COMPLETE') }
    let!(:housing_application) { create(:housing_application, housing_offer: housing_offer, housing_request: housing_request, status: 'TRANSMIS BAILLEUR' ) }
    
    it "should create public shares if request is applicated" do
      hrs = Hermes::HousingRequestShare.find_by(housing_request_id: housing_request.id, lessor_id: lessor.id)
      expect(hrs).not_to be_nil
    end

    it "should create and link housing_request_folder" do
      folders = Hermes::HousingRequestFolder.where(housing_request_id: housing_request.id)
      expect(housing_request.reload.housing_request_folder_id).not_to be_nil
      expect(folders.count).to eq(1)
    end
  end 

  describe "Entity validations" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '12345678901234') }
    let!(:lessor2) { create(:lessor) }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: true, siret: '22345678901234') }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity.siret) }
    let!(:housing_request) { create(:housing_request, is_active: true, main_applicant_id: applicant.id, request_status: 'COMPLETE') }
  
    it "should automaticly validate request" do
      applicant.update(main_employer_siret: entity2.siret)
    
      expect(housing_request.reload.entity_validated).to eq(true)
    end
  
    it "should not automaticly validate request" do
      entity2.update(auto_request_validation: false)
      applicant.update(main_employer_siret: entity2.siret)
    
      expect(housing_request.reload.entity_validated).to eq(false)
    end
  
    it 'should auto set entity_status' do
      hr = FactoryBot.create(:housing_request, is_active: true, main_applicant: applicant)
     # expect(hr.entity_status).to eq(nil)

      hr.update(consent_level: 4)
      # expect(hr.entity_status).to eq('EN COURS DE SAISIE')
    end
  end
  describe "Entity_status workflow" do
    let!(:user) { create(:user) }
    let!(:lessor) { create(:lessor) }
    let!(:entity) { create(:entity, lessor: lessor, siret: '00000000000001') }
    let!(:entity2) { create(:entity, lessor: lessor2, auto_request_validation: false, siret: '00000000000002') }
    let!(:lessor2) { create(:lessor) }
    let!(:applicant) { create(:applicant, user: user, main_employer_siret: entity2.siret, amount_year_minus_two: 200000) }
    let!(:hr) { create(:complete_request, is_active: true, main_applicant: applicant) }
  
    it 'works by default' do 
      expect(hr.reload.entity_status).to eq('A VALIDER')
    end 

    it 'works when entity change autovalidation' do 
      entity2.update(auto_request_validation: true)
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 

    it 'works when entity_validated' do 
      hr.update(entity_validated: true)
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 
    
  end

  describe 'consent_level 1 and entity points' do 
    let!(:entity) { create(:entity, siret: '00000000000001', spent_on_ongoing_requests: 200)}
    let!(:entity2) { create(:entity, siret: '00000000000002', auto_request_validation: false)}
    let!(:entity_user) { create(:user, group_ids: [entity.manager_group_id])}
    let!(:app) { create(:applicant, birth_date: Date.today - 50.years, main_employer_siret: entity.siret, amount_year_minus_two: 11675,  monthly_incomes: {'Salaire': 1345.45}) }
    let!(:hr) { create(:complete_request, main_applicant: app, is_active: true, seeked_localisation_list: [ {codeDepartement: '34'} ],) }
    let!(:entity_form) { Hermes::HousingRequestEntityForm.find_by(housing_request_id: hr.id) }
    let!(:threshold1) do
      create(
        :threshold,
        type: 'PLA LM',
        thresholds_map: {
          'Paris et communes limitrophes': {
            'thresholds_list': [
              13_050,
              21_272,
              27_883,
              30_521,
              36_307,
              40_859
            ],
            'another_person': 4552
          },
        }
      )
    end
    let!(:offer) { create(:housing_offer,           
      rent_with_charges: 234,
      rent_amount: 234,
      typology: 'T2',
      department: '34',
      commercialized: false,
      threshold_type: 'PLA LM',
      offer_status: 'LOGEMENT PUBLIE',
      threshold_subtype: 'Paris et communes limitrophes'
    ) }


    before do
      hr.validate_all_required_document
      hr.update(
        priority_points_indexes: [10]
      )
    end

    it 'has a link to entity' do
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(2)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq('EN COURS')
      expect(hr.entity_id).to eq(entity.id)
      expect(app.readable_by?(entity_user)).to eq true
      expect(hr.readable_by?(entity_user)).to eq true
      expect(Hermes::HousingRequest.where(offer.eligibility_query.json_to_h).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(210)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(210)
    end 

    it 'lose link to entity when consent_level == 1' do 
      entity_form.update(level: 1)
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(1)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq(nil)
      expect(app.main_employer_siret).to eq(entity.siret)
      expect(hr.old_entity_id).to eq(entity.id)
      expect(hr.entity_id).to eq(nil)
      expect(app.readable_by?(entity_user)).to eq false
      expect(hr.readable_by?(entity_user)).to eq false
      expect(Hermes::HousingRequest.where(offer.eligibility_query.json_to_h).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
    end 

    it 'switch back from consent_level == 1' do 
      entity_form.update(level: 1)
      entity_form.update(level: 3)
      hr.reload
      app.reload
      entity.reload
      expect(hr.consent_level).to eq(3)
      expect(hr.entity_id).to eq(entity.id)
      expect(hr.request_status).to eq('COMPLETE')
      expect(hr.entity_status).to eq('EN COURS')
      expect(hr.old_entity_id).to eq(entity.id)
      expect(app.readable_by?(entity_user)).to eq true
      expect(hr.readable_by?(entity_user)).to eq true
      expect(Hermes::HousingRequest.where(offer.eligibility_query.json_to_h).pluck(:id).include?(hr.id)).to eq true
      expect(entity.spent_on_ongoing_requests).to eq(210)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(210)
    end 

    it 'when priority_points_indexes change' do 
      hr.update(priority_points_indexes: [25])
      hr.reload
      app.reload
      entity.reload
      expect(entity.spent_on_ongoing_requests).to eq(225)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(225)
    end 

    it 'when entity_id_changed change' do 
      app.update(main_employer_siret: entity2.siret)
      hr.reload
      app.reload
      entity.reload
      entity2.reload
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
      expect(entity2.spent_on_ongoing_requests).to eq(0)
      expect(entity2.spent_on_completed_requests).to eq(0)
      expect(entity2.spent_on_other).to eq(0)
      expect(entity2.total_spent).to eq(0)
      expect(hr.priority_points_indexes.sum).to eq(0)
      hr.update(entity_validated: true, priority_points_indexes: [10])
      hr.reload
      app.reload
      entity.reload
      entity2.reload
      expect(entity.spent_on_ongoing_requests).to eq(200)
      expect(entity.spent_on_completed_requests).to eq(0)
      expect(entity.spent_on_other).to eq(0)
      expect(entity.total_spent).to eq(200)
      expect(entity2.spent_on_ongoing_requests).to eq(10)
      expect(entity2.spent_on_completed_requests).to eq(0)
      expect(entity2.spent_on_other).to eq(0)
      expect(entity2.total_spent).to eq(10)
    end 

    (Hermes::HousingRequest.remove_entity_stock_status - ['REFUSE']).each do |type|
      it "entity_status #{type}" do 
        type == 'ANNULE' ? hr.update(request_status: 'ANNULEE') : hr.update(entity_status: type) 
        drain_all_jobs
        hr.reload
        app.reload
        entity.reload
        expect(hr.entity_status).to eq(type)
        expect(entity.spent_on_ongoing_requests).to eq(200)
        expect(entity.spent_on_completed_requests).to eq(0)
        expect(entity.spent_on_other).to eq(0)
        expect(entity.total_spent).to eq(200)
        puts "------------------"
        type == 'ANNULE' ? hr.update(request_status: 'COMPLETE') : hr.update(entity_status: 'EN COURS') 
        drain_all_jobs
        hr.reload
        app.reload
        entity.reload
        expect(entity.spent_on_ongoing_requests).to eq(210)
        expect(entity.spent_on_completed_requests).to eq(0)
        expect(entity.spent_on_other).to eq(0)
        expect(entity.total_spent).to eq(210)
      end 
    end 
  end 

  describe 'Coapplicants Validations' do
    let(:main_applicant) { create(:applicant) }
    let(:housing_request) { create(:housing_request, main_applicant: main_applicant, is_active: true) }
    let(:coapplicant) { create(:applicant) }

    it 'can add coapplicants with blank rel_type' do
      housing_request.update(coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.reload.coapplicants.first[:rel_type]).to be_blank
    end

    it 'can not add coapplicants with blank rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: '' }])
      expect(housing_request.errors.any?).to be(true)
      field, error = housing_request.errors.first
      expect(field).to eq(:coapplicants)
      expect(error).to include("rel_type can't be blank for coappllicants")
    end

    it 'can add coapplicants with rel_type if form_status is complete' do
      housing_request.update(form_status: 'complete', coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C' }])
      expect(housing_request.errors.any?).to be(false)
    end
  end

  describe 'typology_range & min_surface' do
    let(:applicant) { create(:applicant, amount_year_minus_two: 50000) }
    let(:coapplicant) { create(:applicant) }
    let(:hr) { create(:housing_request, is_active: true, main_applicant: applicant ) }

    describe 'right for' do 
      pre_mapping = {
        '1 personne seule' => { typology: 1..2, min_surface: 9, less_limited_dependent_nb: 0, coapplicant_nb: 0 },
        '2 personnes' => { typology: 1..3, min_surface: 16, less_limited_dependent_nb: 0, coapplicant_nb: 1 },
        '2 personnes + 1 enfant à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 1, coapplicant_nb: 1 },
        '1 personne seule + 1 enfant à charge' => { typology: 2..3, min_surface: 16, less_limited_dependent_nb: 1, coapplicant_nb: 0 },
        '2 personnes + 2 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 2, coapplicant_nb: 1 },
        '1 personne seule + 2 enfants à charge' => { typology: 2..4, min_surface: 25, less_limited_dependent_nb: 2, coapplicant_nb: 0 },
        '2 personnes + 3 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 3, coapplicant_nb: 1 },
        '1 personne seule + 3 enfants à charge' => { typology: 3..5, min_surface: 34, less_limited_dependent_nb: 3, coapplicant_nb: 0 },
        '2 personnes + 4 enfants à charge' => { typology: 4..7, min_surface: 52, less_limited_dependent_nb: 4, coapplicant_nb: 1 },
        '1 personne seule + 4 enfants à charge' => { typology: 4..6, min_surface: 43, less_limited_dependent_nb: 4, coapplicant_nb: 0 },
        '2 personnes + 5 enfants à charge' => { typology: 5..8, min_surface: 61, less_limited_dependent_nb: 5, coapplicant_nb: 1 },
        '1 personne seule + 5 enfants à charge' => { typology: 5..7, min_surface: 52, less_limited_dependent_nb: 5, coapplicant_nb: 0 },
        '2 personnes + 6 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 6, coapplicant_nb: 1 },
        '1 personne seule + 6 enfants à charge' => { typology: 5..8, min_surface: 61, less_limited_dependent_nb: 6, coapplicant_nb: 0 },
        '2 personnes + 7 enfants à charge' => { typology: 5..10, min_surface: 79, less_limited_dependent_nb: 7, coapplicant_nb: 1 },
        '1 personne seule + 7 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 8 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
        '1 personne seule + 8 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 9 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
        '1 personne seule + 9 enfants à charge' => { typology: 5..9, min_surface: 70, less_limited_dependent_nb: 7, coapplicant_nb: 0 },
        '2 personnes + 10 enfants à charge' => { typology: 5..11, min_surface: 88, less_limited_dependent_nb: 8, coapplicant_nb: 1},
      }

      mappings = pre_mapping.each do |k, a|
        pre_mapping[k][:typology] = (a[:typology]).to_a.map do |i|
          "T#{i.to_s}" 
        end
        pre_mapping[k][:typology_filter] = 
          (pre_mapping[k][:typology] & ['T6', 'T7', 'T8']).empty? ? pre_mapping[k][:typology] : pre_mapping[k][:typology] + ['T6P']
      end

      mappings.each do |k, mapping|
        context k do 
          it 'get the good values' do
            hr.dependents = Array.new(mapping[:less_limited_dependent_nb], { test: 'test'}) 
            hr.coapplicants = mapping[:coapplicant_nb] == 1 ? [ {id: coapplicant.id, rel_type: 'C'}] : [] 
            expect(hr.calculate_typology_range).to eq mapping[:typology]
            expect(hr.calculate_min_surface).to eq mapping[:min_surface]
            hr.typology_range = hr.calculate_typology_range
            expect(hr.typology_filter[:typology][:$in]).to eq mapping[:typology_filter]
          end
        end
      end
      
      it 'unborn_children active' do
        birth_date = (DateTime.now + 150.days).to_s
        hr.update(unborn_children: [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ])
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        expect(hr.main_applicant.eligibility_query.json_to_h[:typology]["$in"]).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
      end

      it 'unborn_children inactive' do
        birth_date = (DateTime.now + 200.days).to_s
        hr.update(unborn_children: [
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
          {"anticipated_birth"=> birth_date, "nb_child_birth"=>"1"},
        ])
        
        expect(hr.calculate_unborn_children_nb).to eq 0
        expect(hr.typology_range).to eq mappings['1 personne seule'][:typology]
        expect(hr.min_surface).to eq mappings['1 personne seule'][:min_surface]
        expect(enqueued_jobs.select {|j| j[:job] == UnbornChildrenJob}.length).to eq 1
        
        Timecop.freeze(DateTime.now + 50.days)
        drain_all_jobs
        expect(hr.calculate_unborn_children_nb).to eq 2
        expect(hr.reload.typology_range).to eq mappings['1 personne seule + 2 enfants à charge'][:typology]
        expect(hr.reload.min_surface).to eq mappings['1 personne seule + 2 enfants à charge'][:min_surface]
        Timecop.return
      end

    end

    describe "inherit birth date from applicant" do
      let!(:applicant) do
        create(
          :applicant,
          birth_date: '2000-01-01'
        )
      end
      let(:housing_request) do
        create(
          :housing_request,
          is_active: true,
          main_applicant: applicant
        )
      end
  
      it 'at creation' do
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end

      it 'changed applicant birth date' do
        applicant.birth_date = "1900-01-01"
        applicant.save
        applicant.reload
        expect(housing_request.requester_birth_date).to eq(applicant.birth_date)
      end
    end

  end

  describe 'calculate category_type for' do 

    context 'one person' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant)}

      it 'as good category' do 
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'one persone with unborn_children' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, unborn_children: [{anticipated_birth: (Date.today + 150.days).to_s, nb_child_birth: 1}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(false)
        expect(hr.reload.category_type).to eq(1)
      end 
    end 

    context 'two people (not young couple)' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(2)
      end 
    end 
    
    context 'one person with handicap' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context 'single_parent with visit' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'D'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context '3 people' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant2) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '1 person with on dependent' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '1 person with on dependent parent' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{coparentality: 'P'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(2)
      end 
    end 

    context 'young couple' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '2 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'} ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context '4 people' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant2) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let(:coapplicant3) { create(:applicant, birth_date: DateTime.now - 30.years )}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}, {id: coapplicant3.id.to_s, rel_type: 'C'}])}
      
      it 'as good category' do 
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 2 dependents + one visit' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}, {rel_type: 'D'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(5)
      end 
    end 

    context '2 person with 2 dependents' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}], main_applicant: applicant, dependents: [{rel_type: 'G'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '2 person with 2 dependents + one visit' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'}], main_applicant: applicant, dependents: [{rel_type: 'D'}, {rel_type: 'enfant à charge'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(false)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '3 people with handicap' do 
      let(:applicant) { create(:applicant)}
      let(:coapplicant) { create(:applicant)}
      let(:coapplicant2) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [ {id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'} ], disability_details: [{lol: 'test'}])}
      
      it 'as good category' do 
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(4)
      end 
    end 

    context '1 person with 3 dependents' do 
      let(:applicant) { create(:applicant)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, dependents: [{rel_type: 'E'}, {rel_type: 'enfant à charge'}, {rel_type: 'garde alternée'}])}
      
      it 'as good category' do 
        expect(hr.reload.single_parent).to eq(true)
        expect(hr.reload.category_type).to eq(5)
      end 
    end 

    ##########################################

    context 'young couple with 1 children' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'E'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with 1 parent' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request, is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], dependents: [{rel_type: 'P'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with handicap' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 20.years)}
      let!(:hr) { create(:housing_request,is_active: true, main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}], disability_details: [{lol: 'test'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.handicap?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 

    context 'young couple with changing date' do 
      let(:applicant) { create(:applicant, birth_date: DateTime.now - 20.years )}
      let(:coapplicant) { create(:applicant, birth_date: DateTime.now - 40.years)}
      let!(:hr) { create(:housing_request, is_active: true,  main_applicant: applicant, coapplicants: [{ id: coapplicant.id.to_s, rel_type: 'C'}] )}
      
      it 'as good category' do 
        expect(hr.reload.young_couple?).to eq(false)
        expect(hr.reload.category_type).to eq(2)
        coapplicant.update(birth_date: DateTime.now - 20.years)
        expect(hr.reload.young_couple?).to eq(true)
        expect(hr.reload.category_type).to eq(3)
      end 
    end 
  end 

  describe 'inheriht reference_tax_revenues from coaplicants' do
    let!(:applicant) {create(:applicant)}
    let!(:coapplicant) {create(:applicant)}
    let!(:coapplicant2) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, coapplicants: [{id: coapplicant.id.to_s, rel_type: 'C'}, {id: coapplicant2.id.to_s, rel_type: 'C'}], is_active: true, form_status: 'invalid', seeked_departments: ['34'])}
    let!(:nud) { create(:nud,housing_request: hr, unique_number: '1')}

    it 'without amount_year_minus_two' do
      expect(hr.reload.reference_tax_revenues).to eq(nil)
    end

    it 'without amount_year_minus_two 0' do
      applicant.update(amount_year_minus_two: 0)
      expect(hr.reload.reference_tax_revenues).to eq(0)
    end

    it 'only applicant amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      expect(hr.reload.reference_tax_revenues).to eq(1000)
    end
    
    it 'all applicants amount_year_minus_two' do
      applicant.update(amount_year_minus_two: 1000)
      coapplicant.update(amount_year_minus_two: 2000)
      expect(hr.reload.reference_tax_revenues).to eq(3000)
      coapplicant2.update(amount_year_minus_two: 500)
      expect(hr.reload.reference_tax_revenues).to eq(3500)
    end

    it 'coapplicant affected after hr creation' do
      coapp = Hermes::Applicant.create(amount_year_minus_two: 800, year_minus_two: Date.today.year - 2)
      hr.update(coapplicants: [{id: coapp.id, rel_type: 'C'}])
      expect(hr.reload.reference_tax_revenues).to eq(800)
    end

    it 'request_status changed' do
      hr.validate_all_required_document
      hr.update(form_status: 'invalid')
      expect(hr.reload.reference_tax_revenues).to eq(nil)
      expect(hr.reload.request_status).to eq("INCOMPLETE")
      
      applicant.update(amount_year_minus_two: 1000)

      coapplicant.update(amount_year_minus_two: 2000)
      coapplicant2.update(amount_year_minus_two: 500)
      hr.update(form_status: 'complete')
      expect(hr.reload.reference_tax_revenues).to eq(3500)
      expect(hr.reload.request_status).to eq("COMPLETE")

      applicant.update(amount_year_minus_two: nil)
      coapplicant.update(amount_year_minus_two: nil)
      coapplicant2.update(amount_year_minus_two: nil)
      expect(hr.reload.reference_tax_revenues).to eq(nil)
      expect(hr.reload.request_status).to eq("INCOMPLETE")

      applicant.update(amount_year_minus_two: 1000)
      expect(hr.reload.reference_tax_revenues).to eq(1000)
      expect(hr.reload.request_status).to eq("INCOMPLETE")
    end
  end


  describe 'housing_channel_keys' do 
    let!(:dalo){ create(:housing_channel, key: 'dalo')}
    let!(:lol){ create(:housing_channel, key: 'lol')}
    let!(:applicant){ create(:applicant)}
    let!(:request1) { create(:housing_request, main_applicant: applicant)}
    let!(:request2) { create(:housing_request, main_applicant: applicant, housing_channel_keys: ['dalo'])}

    it 'create with and without housing_channel_keys' do 
      expect(request1.housing_channel_keys).to eq(['dc'])
      expect(request2.housing_channel_keys).to eq(['dalo', 'dc'])
    end 

    it 'when delete housing_channel_keys' do 
      request2.update(housing_channel_keys: [])
      expect(request2.reload.housing_channel_keys).to eq(['dc'])
    end 

    it 'when add housing_channel_keys' do 
      request1.update(housing_channel_keys: ['dalo'])
      expect(request1.reload.housing_channel_keys).to eq(['dalo', 'dc'])
    end 
  end 

  describe 'update_readers_writers' do
    let!(:entity) {create(:entity, siret: '00000000000001')} 
    let!(:entity_user){ create(:user)}
    let!(:user_right) { create(:user_right, type: 'entity', entity: entity, user: entity_user, right: 'administrator')}
    let!(:applicant){ create(:applicant, user: user)}
    let!(:coapp){ create(:applicant, user: user)}
    let!(:request) { create(:housing_request, main_applicant: applicant, is_active: true, user: user, coapplicants: [{id: coapp.id, rel_type: 'C'}])}

    it 'setup default' do
      readers = []
      writers = [request.user_id.to_s]
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers)
        expect(r.writers).to eq(writers)
        expect(request.readable_by?(entity_user)).to eq(false)
      end
    end 

    it 'when entity change' do
      readers = []
      writers = [request.user_id.to_s]
      readers_with_entity = [ "g_" + entity.group_id]
      writers_with_entity = writers + [entity.supervisor_group_id, entity.manager_group_id].map {|id| "g_" + id.to_s}

      applicant.update(main_employer_siret: entity.siret)
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers_with_entity)
        expect(r.writers).to eq(writers_with_entity)
        expect(request.readable_by?(entity_user)).to eq(true)
      end

      applicant.update(main_employer_siret: nil)
      [applicant.reload, coapp.reload, request.reload].each do |r|
        expect(r.readers).to eq(readers)
        expect(r.writers).to eq(writers)
        expect(request.readable_by?(entity_user)).to eq(false)
      end
    end 
  end 

  describe 'update_request_from_sne update dalo and housing_channel_keys' do
    let!(:dalo){ create(:housing_channel, key: 'dalo')}
    let!(:applicant) {create(:applicant)}
    let!(:hr_with_dalo) {create(:housing_request, main_applicant: applicant)}
    let!(:nud1) { create(:nud,housing_request: hr_with_dalo, unique_number: '1110121047572GDPUB')}
    let!(:hr_without_dalo) {create(:housing_request, main_applicant: applicant)}
    let!(:nud2) { create(:nud,housing_request: hr_without_dalo, unique_number: '1110121021545G1104')}

    it 'works with dalo when empty' do 
      hr_with_dalo.update_request_from_sne(user)
      hr_with_dalo.reload
      expect(hr_with_dalo.dalo).to eq(true)
      expect(hr_with_dalo.housing_channel_keys).to eq([ 'dc', 'dalo'])
    end 

    it 'works with dalo when already set' do 
      hr_with_dalo.update(housing_channel_keys: ['dalo', 'dc'], dalo: true)
      hr_with_dalo.update_request_from_sne(user)
      hr_with_dalo.reload
      expect(hr_with_dalo.dalo).to eq(true)
      expect(hr_with_dalo.housing_channel_keys).to eq(['dalo', 'dc'])
    end 

    it 'works without dalo when empty' do 
      hr_without_dalo.update_request_from_sne(user)
      hr_without_dalo.reload
      expect(hr_without_dalo.dalo).to eq(false)
      expect(hr_without_dalo.housing_channel_keys).to eq(['dc'])
    end 

    it 'works with dalo when already set' do 
      hr_without_dalo.update(housing_channel_keys: ['dalo', 'dc'], dalo: true)
      hr_without_dalo.update_request_from_sne(user)
      hr_without_dalo.reload
      expect(hr_without_dalo.dalo).to eq(false)
      expect(hr_without_dalo.housing_channel_keys).to eq(['dc'])
    end 

    it 'dalo updated when housing_channel_keys change' do 
      expect(hr_with_dalo.reload.dalo).to eq(false)

      hr_with_dalo.update!(housing_channel_keys: ['dalo', 'dc'])
      expect(hr_with_dalo.reload.dalo).to eq(true)

      hr_with_dalo.update!(housing_channel_keys: ['dc'])
      expect(hr_with_dalo.reload.dalo).to eq(false)
    end 

  end 

  describe 'update_request_from_sne dynamic' do 
    let!(:hr) { create(:complete_request, is_active: true, unique_numbers: ['1110121021545G1104'])}

    before do
      Timecop.freeze(Time.local(2021, 2, 1, 0, 0))
      hr.update_request_from_sne(nil)
      hr.validate_all_required_document
      Timecop.return
    end 

    it 'default' do 
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier validé')
      expect(hr.housing_request_folder.application_status).to eq('Dossier validé')
      expect(hr.reference_tax_revenues).to eq(22000)
      expect(hr.request_status).to eq('COMPLETE')
    end 

    it 'with config same year' do 
      date = Date.new(2021, 4, 1,)
      HERMES_CONFIG['change_minus_two_date'] = date + 10.days
      HERMES_CONFIG['change_minus_two_year'] = date.year - 1
      Timecop.freeze(date)
      hr.update_request_from_sne(nil)
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.reference_tax_revenues).to eq(22000)
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.application_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.get_not_declared.all? { |type| type.include?('Bulletin')}).to eq(true)
      Timecop.return
    end 
    
    it 'with config next year' do 
      date = Date.new(2021, 4, 1,)
      HERMES_CONFIG['change_minus_two_date'] = date - 10.days
      HERMES_CONFIG['change_minus_two_year'] = date.year - 1
      Timecop.freeze(date)
      hr.update_request_from_sne(nil)
      hr.reload
      hr.housing_request_folder.reload
      expect(hr.reference_tax_revenues).to eq(22500)
      expect(hr.housing_request_folder.consultation_status).to eq('Dossier incomplet')
      expect(hr.housing_request_folder.application_status).to eq('Dossier incomplet')
      Timecop.return

      # Clear config
      HERMES_CONFIG['change_minus_two_date'] = nil
      HERMES_CONFIG['change_minus_two_year'] = nil
    end 

  end 

  describe 'update_request_from_sne with multiple dependents monthly incomes' do
    let!(:applicant) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant)}
    let!(:nud1) { create(:nud,housing_request: hr, unique_number: '1110121047572GDPUB')}

    it 'works' do 
      hr.update_request_from_sne(user)
      hr.reload
      expect(hr.dependents.first['monthly_incomes'].length).to eq(15)
      expect(hr.dependents.first['monthly_charges'].length).to eq(2)
      expect(hr.dependents_monthly_revenues).to eq(790)
    end 
  end 

  describe 'score' do 
    let!(:entity) {create(:entity, siret: '00000000000001', auto_request_validation: false)}
    let!(:hr) {create(:complete_request)}
    let!(:offer) {create(:housing_offer)}
    let!(:ha) { create(:housing_application, status: 'EN ATTENTE', housing_request: hr, housing_offer: offer)}

    it 'work with minimium info' do 
      hr.reload
      expect(hr.profile_score).to eq 0
      expect(hr.entity_points).to eq 0
      expect(hr.motivation_points).to eq 0
      expect(hr.seniority_points).to eq 0
      expect(hr.under_30_year_points).to eq 0
      expect(ha.profile_score).to eq hr.profile_score
    end
    
    it 'update ha score when score change' do 
      expect(ha.reload.profile_score).to eq 0
      hr.update(sne_created_at: DateTime.now - 30.months)
      expect(hr.reload.profile_score).to eq 30
      expect(ha.reload.profile_score).to eq 30
    end 
    
    it 'when priority_points changed' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(priority_points_indexes: [100])
      expect(hr.profile_score).to eq 100
      expect(hr.entity_points).to eq 100
    end 

    it 'when entity_validated changed' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(entity_validated: true)
      expect(hr.profile_score).to eq 240
      expect(hr.entity_points).to eq 240
    end 

    it 'when auto_request_validation changed' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      entity.update(auto_request_validation: true)
      hr.reload
      expect(hr.profile_score).to eq 240
      expect(hr.entity_points).to eq 240
    end 

    it 'when motivations changed' do 
      hr.update(motivations: ['19'])
      hr.reload
      expect(hr.profile_score).to eq 70
      expect(hr.motivation_points).to eq 70
    end 

    it 'when motivations changed' do 
      hr.update(motivations: ['19'])
      hr.reload
      expect(hr.profile_score).to eq 70
      expect(hr.motivation_points).to eq 70
    end 

    it 'when sne_created_at changed' do 
      hr.update(sne_created_at: DateTime.now - 30.months)
      hr.reload
      expect(hr.profile_score).to eq 30
      expect(hr.seniority_points).to eq 30
    end 

    it 'when applicant birth_date changed' do 
      birth_date = Date.today - 20.years
      hr.main_applicant.update(birth_date: birth_date)
      hr.reload
      expect(hr.requester_birth_date).to eq birth_date
      expect(hr.profile_score).to eq 70
      expect(hr.under_30_year_points).to eq 70
    end 

    it 'when not in progress and not already_had_score' do 
      hr.update(nud_ids: [], unique_numbers: [], already_had_score: false)
      hr.update(motivations: ['19'])
      expect(hr.reload.request_status).to eq "EN ATTENTE DE NUD"

      expect(hr.reload.profile_score).to eq 0
      expect(ha.reload.profile_score).to eq 0
    end 

    it 'when not in progress and already_had_score' do 
      hr.update(nud_ids: [], unique_numbers: [])
      hr.update(motivations: ['19'])

      expect(hr.reload.profile_score).to eq 70.0
      expect(ha.reload.profile_score).to eq 70.0
    end 

    it 'when attente de piece' do 
      hr.housing_request_folder.required_documents.first.update(status: "Pièce manquante", declared: false)
      hr.update(sne_created_at: DateTime.now - 30.months)
      expect(hr.reload.request_status).to eq('EN ATTENTE DE PIECES')
      expect(hr.reload.profile_score).to eq 30
    end   
    
    it 'when bonus_point changed' do 
      hr.update(bonus_points: 13)
      expect(hr.reload.profile_score).to eq 13
    end 
  end 

  describe 'calculate_next_seniority_updated_date' do 
    let!(:applicant) {create(:applicant)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, is_active: true)}

    [
      {
        title: 'for same day',
        current_date: DateTime.new(2020, 1, 2),
        sne_created_at_date: DateTime.new(2019, 7, 2),
        expected_date: DateTime.new(2020, 2, 3),
        expected_seniority_point: 6,
      },      
      {
        title: 'for same month date',
        current_date: DateTime.new(2020, 1, 2),
        sne_created_at_date: DateTime.new(2019, 7, 15),
        expected_date: DateTime.new(2020, 1, 16),
        expected_seniority_point: 5,
      },
      {
        title: 'for next month',
        current_date: DateTime.new(2020, 1, 22),
        sne_created_at_date: DateTime.new(2019, 7, 15),
        expected_date: DateTime.new(2020, 2, 16),
        expected_seniority_point: 6,
      },
      {
        title: 'for not existing date',
        current_date: DateTime.new(2020, 2, 2),
        sne_created_at_date: DateTime.new(2019, 7, 30),
        expected_date: DateTime.new(2020, 3, 1),
        expected_seniority_point: 6,
      },
    ].each do |d|
      it d[:title] do 
        Timecop.freeze(d[:current_date])
        hr.set(sne_created_at: d[:sne_created_at_date])
        expect(hr.calculate_next_seniority_updated_date).to eq(d[:expected_date].utc)
        expect(hr.calculate_seniority_points).to eq(d[:expected_seniority_point])
        Timecop.return
        Timecop.freeze(d[:expected_date])
        expect(hr.calculate_seniority_points).to eq(d[:expected_seniority_point] + 1)
        Timecop.return
      end 
    end 
  end 

  describe 'update_renewal_values' do 
    let!(:created_at) { Date.new(2020, 1, 1)}
    let!(:applicant) {create(:applicant,)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, created_at: created_at)}

    it 'works without sne_created_at' do 
      expect(hr.reload.next_sne_renewal_date).to eq(created_at + 1.year)
    end 

    it 'works with sne_created_at' do 
      sne_created_at = Date.new(2020, 4, 1)
      expect do
        hr.update(sne_created_at: sne_created_at)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(sne_created_at + 1.year)
    end 

    it 'works when last_sne_renewal_date changed' do 
      last_sne_renewal_date = Date.new(2020, 6, 1)
      expect do
        hr.update(last_sne_renewal_date: last_sne_renewal_date)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(last_sne_renewal_date + 1.year)
    end 

    it 'update renewal bolean' do 
      last_sne_renewal_date = Date.today - 350.days
      
      hr.update(last_sne_renewal_date: Date.today - 6.months)
      expect(hr.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr.reload.need_to_be_renewed).to eq(false)

      hr.update(last_sne_renewal_date: last_sne_renewal_date)
      expect(hr.reload.soon_need_to_be_renewed).to eq(true)
      expect(hr.reload.need_to_be_renewed).to eq(false)
 
      hr.update(last_sne_renewal_date: last_sne_renewal_date - 1.month)
      expect(hr.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr.reload.need_to_be_renewed).to eq(true)
    end

    it 'force when next_renewal_date_set_by_sne' do 
      expect do
        hr.update(next_renewal_date_set_by_sne: created_at + 5.days)
      end.to change { hr.reload.next_sne_renewal_date }.from(created_at + 1.year).to(created_at + 5.days)
    end 

  end 

  describe 'update_renewed_booleans' do 
    let!(:today) { Date.today}
    let!(:applicant1) {create(:applicant)}
    let!(:hr1) {create(:housing_request, main_applicant: applicant1, created_at: today)}
    let!(:applicant2) {create(:applicant)}
    let!(:hr2) {create(:housing_request, main_applicant: applicant2, created_at: today - 20.days)}
    let!(:applicant3) {create(:applicant)}
    let!(:hr3) {create(:housing_request, main_applicant: applicant3, created_at: today - 35.days)}


    it 'works' do 
      expect(hr1.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr1.reload.need_to_be_renewed).to eq(false)
      expect(hr2.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr2.reload.need_to_be_renewed).to eq(false)
      expect(hr3.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr3.reload.need_to_be_renewed).to eq(false)

      Timecop.freeze(today + 11.month - 1.days)
      Hermes::HousingRequest.update_renewed_booleans
      
      expect(hr1.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr1.reload.need_to_be_renewed).to eq(false)
      expect(hr2.reload.soon_need_to_be_renewed).to eq(true)
      expect(hr2.reload.need_to_be_renewed).to eq(false)
      expect(hr3.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr3.reload.need_to_be_renewed).to eq(true)

      Timecop.freeze(today + 12.month)
      Hermes::HousingRequest.update_renewed_booleans
      expect(hr2.reload.soon_need_to_be_renewed).to eq(false)
      expect(hr2.reload.need_to_be_renewed).to eq(true)
      Timecop.return

    end 
  end 

  describe 'check common validations' do 
   
    let!(:applicant) {create(:applicant,)}
    let!(:applicant2) {create(:applicant,)}
    let!(:hr) {create(:housing_request, main_applicant: applicant, is_active: true)}
    let!(:entity) {create(:entity, siret: '12345678901234')}
   
    it 'has standard validation errors' do
      applicant.first_name = nil
      applicant.birth_date = nil
      applicant.nationality = nil
      applicant.civility = nil
      applicant.work_contract_type = nil
      applicant.email = nil
      applicant.actual_housing_details['is_house_owner'] = nil
      applicant.postal_details = {etranger: false}
      hr.reload
      hr.seeked_types = []
      hr.seeked_localisation_list = []
      hr.widen_request = nil
      applicant.save
      hr.save
      errors_lines = hr.form_sne_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} + hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]}
      [3,5,6,7,8,9,12,27,36,37,38,39].each do |val|
        puts "sdqdq : #{val}" unless errors_lines.include?(val.to_s) 
        expect(errors_lines.include?(val.to_s)).to eq(true)
      end
    end
    it 'has standard validation errors' do
      applicant.first_name = nil
      applicant.birth_date = nil
      applicant.nationality = nil
      applicant.email = nil
  
      applicant.actual_housing_details = {
        'foreign_address' => "true",
        'is_house_owner' => nil}
      hr.reload
      hr.seeked_types = []
      hr.widen_request = true
      hr.seeked_localisation_list = []
      hr.dependents = [ {
        "last_name": "Nestle ",
        "first_name": "Concentre",
        "birth_date": "2015-05-10",
        "rel_type": "E",
        "sex": "M"
    },
    {
        "last_name": "Nestle ",
        "first_name": "Poudre",
        "birth_date": "2019-05-10",
        "rel_type": "E",
        "sex": "F"
    },
    {
        "last_name": "Nestle ",
        "first_name": "Fraise",
        "birth_date": "1970-05-10",
        "rel_type": "P",
        "sex": "F"
    },
    {
        "last_name": "Nestle",
        "first_name": "Cerise",
        "birth_date": "2010-05-10",
        "coparentality": "G",
        "sex": "F"
    },
    {
        "last_name": "Nestle",
        "first_name": "Tube",
        "birth_date": "2015-05-10",
        "coparentality": "D",
        "sex": "M"
    }]
      applicant.save
      hr.save
      hr.reload
      errors_lines = hr.form_sne_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} + hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]}
      [16,17,18,19, 35].each do |val|
        expect(errors_lines.include?(val.to_s)).to eq(true)
      end
    end
    it 'has standard validation errors' do
      applicant.amount_year_minus_two = 2000
      applicant.main_employer_name = '12345678901234'
      applicant.save
      hr.reload
      hr.widen_request = true
      hr.save
      errors_lines = hr.form_sne_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} + hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]}
      [35,38, 41].each do |val|
        if errors_lines.include?(val.to_s)
          puts "è" * 100  
          puts val 
        end 
        expect(errors_lines.include?(val.to_s)).to eq(false)
      end
    end
    it 'has standard validation errors' do
      applicant.work_contract_type = "CDI"
      applicant.main_employer_name = nil
      applicant.save
      
      hr.reload
      applicant.reload

      hr.applicant_used_for_siret = applicant.id.to_s
      hr.widen_request = true
      hr.save

      errors_lines = hr.form_sne_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} + hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]}
   
      [41].each do |val|
          expect(errors_lines.include?(val.to_s)).to eq(true)
      end
    end

    it 'when no main_employer_name and AGP' do 
      # maybe need to check when applicant changed after housing_request
      applicant2.update(work_contract_type: "AGP", main_employer_name: nil)
      hr.update(applicant_used_for_siret: applicant2.id, coapplicants: [{id: applicant2.id.to_s, rel_type: 'C'}])
      expect(hr.reload.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} ).to include('41')
    end 

    it 'check_consent_level' do 
      expect(hr.reload.entity_id).to be_blank
      expect(hr.consent_level).to be_blank
      expect(hr.reload.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} ).not_to include('43')

      hr.get_applicant_used_for_siret.update(main_employer_siret: entity.siret, work_contract_type: "CDI")
      expect(hr.reload.entity_id).not_to be_blank
      expect(hr.consent_level).to be_blank
      expect(hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} ).to include('43')

      hr.update(consent_level: 3)
      expect(hr.reload.entity_id).not_to be_blank
      expect(hr.consent_level).not_to be_blank
      expect(hr.form_front_validation_errors.map { |e| e["csv_rule_to_check_line_to_check"]} ).not_to include('43')
    end 
  end

  describe 'update_entity_relationship' do 
    let!(:entity1) {create(:entity, siret: '10000000000000', siren: '100000000')}
    let!(:entity11) {create(:entity, siret: '10000000000001', siren: '100000000', parent_id: entity1.id)}
    let!(:entity2) {create(:entity, siret: '20000000000000', siren: '200000000')}  
    let!(:entity3) {create(:entity, siret: '30000000000000', siren: '300000000')}  
    let!(:entity31) {create(:entity, siret: '30000000000001', siren: '300000001')}  
    let!(:entity32) {create(:entity, siret: '30000000000002', siren: '300000001')}  

    let!(:entity_user){ create(:user)}
    let!(:applicant){ create(:applicant, main_employer_siret: entity1.siret)}
    let!(:coapp){ create(:applicant)}
    let!(:entity_code){ create(:entity_code, entity_id: entity2.id, code: 'ENTITY2' )}
    let!(:hr) { create(:housing_request, main_applicant: applicant, is_active: true, user: user, coapplicants: [{id: coapp.id.to_s, rel_type: 'C'}], consent_level: 2)}

    it 'default' do 
      expect(hr.reload.entity_id).to eq(entity1.id)
      expect(hr.reload.consent_level).to eq(2)
    end 

    it 'when last_entity_code' do 
      hr.update(last_entity_code: entity_code.code)
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
      hr.update(last_entity_code: nil)
      expect(hr.reload.entity_id).to eq(entity1.id)
    end 

    it 'when last_entity_code changed from consent level 1' do
      hr.update(consent_level: 1)
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(1)
      hr.update(last_entity_code: entity_code.code)
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when main_employer_siret change' do 
      applicant.update(main_employer_siret: entity3.siret)
      expect(hr.reload.entity_id).to eq(entity3.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when coapplicant main_employer_siret change' do 
      hr.update(applicant_used_for_siret: coapp.id)
      coapp.update(main_employer_siret: entity3.siret)
      expect(hr.reload.entity_id).to eq(entity3.id)
      expect(hr.reload.consent_level).to eq(nil)
      expect(hr.applicant_used_for_siret_firstname).to eq(coapp.first_name)
      coapp.first_name = "dudule"
      coapp.save
      expect(hr.reload.applicant_used_for_siret_firstname).to eq("dudule")
    end 

    it 'when should use siren' do
      applicant.update(main_employer_siret: '20000000000001')
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when siren not work' do
      applicant.update(main_employer_siret: '30000000100000')
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when siret not 14 length' do
      applicant.update(main_employer_siret: '2000000000000')
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when one is_master' do 
      entity2.update(siren: entity1.siren, siret: entity1.siret, is_master: true)
      applicant.update(main_employer_siret: entity1.siren + '12345' )
      expect(hr.reload.entity_id).to eq(entity2.id)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when none is_master' do 
      entity2.update(siren: entity1.siren, siret: entity1.siret, is_master: false)
      applicant.update(main_employer_siret: entity1.siren + '12345' )
      expect(hr.reload.entity_id).to eq(nil)
      expect(hr.reload.consent_level).to eq(nil)
    end 

    it 'when different level' do 
      applicant.update(main_employer_siret: entity1.siren + '12345' )
      expect(hr.reload.entity_id).to eq(entity1.id)
      expect(hr.reload.consent_level).to eq(nil)

    end 
  end
  
  describe 'clear_attributes' do 
    let!(:applicant) {create(:applicant, main_employer_name: 'Test')}
    let!(:hr) {create(:housing_request, main_applicant: applicant, is_active: true)}

    it 'works' do 
      hr.update(is_active: false)
      expect(applicant.reload.main_employer_name).to eq(nil)
    end 
  end 

  describe 'update_applicant_used_for_siret_info' do 
    let!(:app1) { create(:applicant, first_name: 'One', last_name: 'Doe', email: 'one@test.com')}
    let!(:app2) { create(:applicant, first_name: 'Two', last_name: 'Doe', email: 'two@test.com')}
    let!(:hr) {create(:housing_request, is_active: true, main_applicant: app1, coapplicants: [{id: app2.id.to_s, rel_type: 'C'}])}

    it 'default' do 
      expect(hr.applicant_used_for_siret_firstname).to eq('One')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('One Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app1.birth_date)
    end 

    it 'when main_applicant email change' do 
      expect {
        app1.update(email: 'onebis@test.com')
      }.to change { hr.reload.applicant_used_for_siret_email }.from('one@test.com').to('onebis@test.com')
    end 

    it 'when applicant_used_for_siret change' do 
      hr.update(applicant_used_for_siret: app2.id)
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('Two')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('Two Doe')
      expect(hr.applicant_used_for_siret_email).to eq('two@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app2.birth_date)
    end 

    it 'when applicant_used_for_siret and email missing' do 
      hr.update(applicant_used_for_siret: app2.id)
      app2.update(email: nil)
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('Two')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('Two Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app2.birth_date)
    end 

    it 'when applicant_used_for_siret and email missing' do 
      app2.update(email: "twobis@test.com")
      hr.reload
      expect(hr.applicant_used_for_siret_firstname).to eq('One')
      expect(hr.applicant_used_for_siret_lastname).to eq('Doe')
      expect(hr.applicant_used_for_siret_fullname).to eq('One Doe')
      expect(hr.applicant_used_for_siret_email).to eq('one@test.com')
      expect(hr.applicant_used_for_siret_birth_date).to eq(app1.birth_date)
    end 
  end 

  describe 'entity_status workflow' do 
    let!(:hr) { create(:complete_request, user: user)}
    let!(:offer) { create(:housing_offer, district: 'Paris', postal_code: '75000')}
    let!(:entity) { create(:entity, siret: '12345678901234')}

    it 'no entity_id' do 
      expect(hr.reload.entity_status).to eq(nil)
      
      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq(nil)

      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq(nil)
      expect(hr.request_status).to eq('SATISFAITE')
      expect(hr.attributed_housing_district).to eq('Paris')
      expect(hr.attributed_housing_district_postal_code).to eq('75000')
    end

    it 'with cancelled hr' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(form_status: 'cancelled')
      expect(hr.reload.request_status).to eq('ANNULEE')
      expect(hr.reload.entity_status).to eq('ANNULE')
    end 

    it 'with entity_id' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('EN COURS')

      entity.update(auto_request_validation: false)
      hr.update(entity_validated: false)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.update(entity_validated: true)
      expect(hr.reload.entity_status).to eq('EN COURS')
    end 

    it 'with REFUSE' do 
      entity.update(auto_request_validation: false )
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.update(entity_refused: true)
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('REFUSE')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('REFUSE')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('REFUSE')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 

    it 'with A VALIDER' do 
      entity.update(auto_request_validation: false)
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('A VALIDER')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('A VALIDER')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('A VALIDER')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 

    it 'with EN COURS' do 
      hr.main_applicant.update(main_employer_siret: entity.siret)
      hr.reload.update(consent_level: 2)
      expect(hr.reload.entity_status).to eq('EN COURS')

      ha = Hermes::HousingApplication.create(status: 'EN ATTENTE', housing_offer: offer, housing_request: hr )
      expect(hr.reload.entity_status).to eq('EN COURS')

      ha.update!(status: 'TRANSMIS BAILLEUR')
      expect(hr.reload.entity_status).to eq('DOSSIER PRESENTE')

      ha.update!(status: 'NON RETENUE')
      expect(hr.reload.entity_status).to eq('EN COURS')

      hr.set(active_application_id: ha.id)
      ha.update!(status: 'RETENUE EN CAL')
      expect(hr.reload.entity_status).to eq('DOSSIER ACCEPTE')

      ha.update!(status: 'LOGE')
      expect(hr.reload.entity_status).to eq('SALARIE LOGE')
      expect(hr.reload.request_status).to eq('SATISFAITE')
    end 
  end 

  describe 'entity_refused' do 
    let!(:entity) { create(:entity, siret: '12345678901234')}
    let!(:app) { create(:applicant, main_employer_siret: entity.siret)}
    let!(:hr) {create(:housing_request, main_applicant: app)}

    it 'default' do 
      expect(hr.reload.entity_refused).to eq(false)
      expect(hr.reload.entity_status).not_to eq('REFUSE')
    end 

    it 'when true' do 
      hr.update(entity_refused: true)
      expect(hr.reload.entity_status).to eq('REFUSE')
    end 

    it 'force entity_valided' do
      hr.set(entity_validated: true)
      hr.update(entity_refused: true)
      expect(hr.reload.entity_validated).to eq(false)
    end 
  end 


  describe 'only_public eligibility' do 
    let!(:hr) { create(:complete_request)}
    let!(:app) { hr.main_applicant }
    subject { app.reload.eligibility_query.json_to_h }

    it 'default' do 
      expect(subject).not_to eq({is_public: true})
    end 

    it 'only_public when is inactive' do 
      hr.update(is_active: false)
      expect(subject).to eq({is_public: true})
    end 

    it 'not only_public when reference_tax_revenues 0' do 
      hr.update(reference_tax_revenues: 0)
      expect(subject).not_to eq({is_public: true})
    end 

    it 'only_public when reference_tax_revenues nil' do 
      hr.update(reference_tax_revenues: nil)
      expect(subject).to eq({is_public: true})
    end 

    it 'not only_public when monthly_revenues 0' do 
      hr.update(monthly_revenues: 0)
      expect(subject).not_to eq({is_public: true})
    end 

    it 'only_public when monthly_revenues nil' do 
      hr.update(monthly_revenues: nil)
      expect(subject).to eq({is_public: true})
    end 

  end 
end
